# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-datacat entity's classes"""

from os import path

import rdflib

from cubicweb.predicates import (
    has_related_entities,
    is_instance,
)
from cubicweb.entities import (
    adapters,
    AnyEntity,
)

from cubes.datacat import (
    register_catalog_rdf_mapping,
    register_contact_point_rdf_mapping,
    register_dataset_rdf_mapping,
    register_distribution_rdf_mapping,
    register_periodoftime_rdf_mapping,
    register_publisher_rdf_mapping,
)
from cubes.file.entities import FileIDownloadableAdapter
from cubes.skos.entities import AbstractRDFAdapter
from cubes.skos import rdfio


class Agent(AnyEntity):
    __regid__ = 'Agent'

    def dc_title(self):
        """Return a dc_title built from name and email"""
        title = self.name
        if self.email:
            title += ' <%s>' % self.email
        return title


class Dataset(AnyEntity):
    __regid__ = 'Dataset'

    def dc_title(self):
        """dc_title is either the title or the identifier"""
        return self.title or self.identifier

    @property
    def contact_point(self):
        """Contact point of this Dataset"""
        if self.dataset_contact_point:
            return self.dataset_contact_point[0]


class PeriodOfTime(AnyEntity):
    __regid__ = 'PeriodOfTime'

    def dc_title(self):
        return u'{} - {}'.format(
            self.printable_value('start'), self.printable_value('end'))


class Distribution(AnyEntity):
    __regid__ = 'Distribution'

    def dc_title(self):
        if self.title:
            return self.title
        elif self.reverse_file_distribution:
            return self.reverse_file_distribution[0].dc_title()
        else:
            return u'{0} #{1}'.format(self.cw_etype, self.eid)


class DownloadableDistributionFile(FileIDownloadableAdapter):

    __select__ = (FileIDownloadableAdapter.__select__
                  & has_related_entities('file_distribution', role='subject'))

    def download_file_name(self):
        name, ext = path.splitext(self.entity.data_name)
        return name[:50] + ext


class DownloadableDistribution(adapters.IDownloadableAdapter):
    """IDownloadable adapter for Distribution entities related to a File."""

    __select__ = (adapters.IDownloadableAdapter.__select__ &
                  has_related_entities('file_distribution', role='object'))

    @property
    def downloadable_file(self):
        """The IDownloadable adapted distribution file."""
        dfile = self.entity.reverse_file_distribution[0]
        return dfile.cw_adapt_to('IDownloadable')

    def download_url(self, **kwargs):
        name = self._cw.url_quote(self.download_file_name())
        path = '%s/raw/%s' % (self.entity.rest_path(), name)
        return self._cw.build_url(path, **kwargs)

    def download_content_type(self):
        return self.downloadable_file.download_content_type()

    def download_encoding(self):
        return self.downloadable_file.download_encoding()

    def download_file_name(self):
        return self.downloadable_file.download_file_name()

    def download_data(self):
        return self.downloadable_file.download_data()


class ResourceFeed(AnyEntity):
    __regid__ = 'ResourceFeed'

    @property
    def dataset(self):
        """The Dataset using this ResourceFeed."""
        return self.resource_feed_of[0]

    @property
    def data_format(self):
        return self.media_type[0].label()


class TransformationScript(AnyEntity):
    __regid__ = 'TransformationScript'

    @property
    def output_format(self):
        if self.output_media_type:
            return self.output_media_type[0].label()

    @property
    def transformation_sequence(self):
        return self.reverse_step_script[0].in_sequence[0]


#
# Adapters
#

# Export to RDF

def license_rdf_uri(license):
    """Return the URI to be used as license RDF identifier."""
    exact_matches = license.exact_match
    if exact_matches and exact_matches[0].cw_etype == 'ExternalUri':
        return exact_matches[0].uri
    else:
        return rdfio.permanent_url(license)


def fill_graph_with(entity, graph, adapterid):
    """Fill a `graph` using `entity` adaped as `adapterid`."""
    filler = entity.cw_adapt_to(adapterid)
    assert filler, '{0} not adaptable as {1}'.format(entity, adapterid)
    filler.fill(graph)


def add_literal_to_graph(subject_s_uri, property_s_uri, value, graph):
    """Add the following triple to RDF graph: ``(subject_s_uri, property_s_uri, value).``"""
    if value is not None:
        graph.add(graph.uri(subject_s_uri), graph.uri(property_s_uri), value)


def add_uri_to_graph(subject_s_uri, property_s_uri, object_s_uri, graph):
    """Add the following triple to RDF graph: ``(subject_s_uri, property_s_uri, object_s_uri)``."""
    graph.add(graph.uri(subject_s_uri), graph.uri(property_s_uri), graph.uri(object_s_uri))


def add_license_to_graph(license, subject_uri, graph, reg):
    """Add license information to RDF graph for `subject_uri`."""
    reg.register_prefix('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
    license_uri = license_rdf_uri(license)
    add_uri_to_graph(subject_uri, reg.normalize_uri('dcterms:license'), license_uri, graph)
    add_uri_to_graph(license_uri, reg.normalize_uri('rdf:type'),
                     reg.normalize_uri('dcterms:LicenseDocument'), graph)
    add_literal_to_graph(license_uri, reg.normalize_uri('dcterms:title'),
                         license.label(language_code=license._cw.lang), graph)
    if license.license_type:
        concept = license.license_type[0]
        add_concept_to_graph(concept, license_uri, 'dcterms:type', graph, reg)


def add_checksum_to_graph(checksum_value, subject_uri, graph, reg):
    """Add a spdx:Checksum related to `subject_uri` through spdx:checksum.

    spdx:algorithm is fixed to checksumAlgorithm_sha1 as this is currently the
    only supported value per DCAT-AP.
    """
    reg.register_prefix('spdx', 'http://spdx.org/rdf/terms#')
    # Set a predictable but unique identifier for spdx:Checksum blank node.
    checksum = rdflib.BNode('/'.join([subject_uri, checksum_value]))
    add_uri_to_graph(subject_uri, reg.normalize_uri('spdx:checksum'),
                     checksum, graph)
    add_uri_to_graph(checksum, reg.normalize_uri('rdf:type'),
                     reg.normalize_uri('spdx:Checksum'), graph)
    add_literal_to_graph(checksum, reg.normalize_uri('spdx:checksumValue'),
                         checksum_value, graph)
    add_uri_to_graph(checksum, reg.normalize_uri('spdx:algorithm'),
                     reg.normalize_uri('spdx:checksumAlgorithm_sha1'),
                     graph)


def graph_add_concept_as(graph, reg, concept, subject_uri, rdf_property, as_type):
    """Add `concept` as an rdf:type `as_type` to graph related to
    `subject_uri` through `rdf_property`.
    """
    reg.register_prefix('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
    object_uri = concept.cwuri
    add_uri_to_graph(subject_uri, reg.normalize_uri(rdf_property), object_uri,
                     graph)
    add_uri_to_graph(object_uri, reg.normalize_uri('rdf:type'),
                     reg.normalize_uri(as_type), graph)
    add_literal_to_graph(object_uri, reg.normalize_uri('rdf:value'),
                         concept.label(), graph)


def add_concept_scheme_to_graph(concept_scheme, graph, reg):
    """Add concept scheme to RDF graph related to `subject_uri` with 'relation_uri'."""
    reg.register_prefix('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
    reg.register_prefix('skos', 'http://www.w3.org/2004/02/skos/core#')
    reg.register_prefix('dcterms', 'http://purl.org/dc/terms/')
    scheme_uri = concept_scheme.cwuri
    add_uri_to_graph(scheme_uri, reg.normalize_uri('rdf:type'),
                     reg.normalize_uri('skos:ConceptScheme'), graph)
    add_literal_to_graph(scheme_uri, reg.normalize_uri('dcterms:title'), concept_scheme.title,
                         graph)


def add_concept_to_graph(concept, subject_uri, relation_uri, graph, reg):
    """Add concept to RDF graph related to `subject_uri` with 'relation_uri'."""
    reg.register_prefix('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
    reg.register_prefix('skos', 'http://www.w3.org/2004/02/skos/core#')
    concept_uri = concept.cwuri
    add_uri_to_graph(subject_uri, reg.normalize_uri(relation_uri), concept_uri, graph)
    add_uri_to_graph(concept_uri, reg.normalize_uri('rdf:type'), reg.normalize_uri('skos:Concept'),
                     graph)
    add_uri_to_graph(concept_uri, reg.normalize_uri('skos:inScheme'), concept.in_scheme[0].cwuri,
                     graph)
    for label in concept.preferred_label:
        add_literal_to_graph(concept_uri, reg.normalize_uri('skos:prefLabel'),
                             rdfio.unicode_with_language(label.label, label.language_code), graph)


def add_concept_literal_to_graph(concept, subject_uri, relation_uri, graph, reg):
    """Add concept's label to RDF graph related to `subject_uri` with 'relation_uri'."""
    reg.register_prefix('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
    reg.register_prefix('skos', 'http://www.w3.org/2004/02/skos/core#')
    label = concept.label(language_code=concept._cw.lang)
    add_literal_to_graph(subject_uri, reg.normalize_uri(relation_uri), label, graph)


class DataCatalogRDFAdapter(AbstractRDFAdapter):
    """Adapt DataCatalog entities to RDF using DCAT vocabulary."""
    __regid__ = 'RDFPrimary'
    __select__ = AbstractRDFAdapter.__select__ & is_instance('DataCatalog')

    register_rdf_mapping = staticmethod(register_catalog_rdf_mapping)

    def fill(self, graph):
        super(DataCatalogRDFAdapter, self).fill(graph)
        reg = self.registry
        reg.register_prefix('foaf', 'http://xmlns.com/foaf/0.1/')
        catalog_uri = rdfio.permanent_url(self.entity)
        for agent in self.entity.catalog_publisher:
            fill_graph_with(agent, graph, self.__regid__)
        # Export license.
        for license in self.entity.license:
            add_license_to_graph(license, catalog_uri, graph, reg)
        # Export homepage attribute as an RDF resource
        if self.entity.homepage:
            add_uri_to_graph(catalog_uri, reg.normalize_uri('foaf:homepage'), self.entity.homepage,
                             graph)
        # Export languages
        for language in self.entity.language:
            fill_graph_with(language, graph, self.__regid__)
            graph_add_concept_as(graph, reg, language, catalog_uri,
                                 'dcterms:language', 'dcterms:LinguisticSystem')
        # Export concept scheme
        for scheme in self.entity.theme_taxonomy:
            add_concept_scheme_to_graph(scheme, graph, reg)
            add_uri_to_graph(catalog_uri, reg.normalize_uri('dcat:themeTaxonomy'), scheme.cwuri,
                             graph)
        # Export locations
        for location in self.entity.spatial_coverage:
            fill_graph_with(location, graph, self.__regid__)
            graph_add_concept_as(graph, reg, location, catalog_uri,
                                 'dcterms:spatial', 'dcterms:Location')


class DatasetRDFAdapter(AbstractRDFAdapter):
    """Adapt Dataset entities to RDF using DCAT vocabulary."""
    __regid__ = 'RDFPrimary'
    __select__ = AbstractRDFAdapter.__select__ & is_instance('Dataset')

    register_rdf_mapping = staticmethod(register_dataset_rdf_mapping)

    def fill(self, graph):
        super(DatasetRDFAdapter, self).fill(graph)
        reg = self.registry
        dataset_uri = rdfio.permanent_url(self.entity)
        # Export comma-separated keyword attribute as multiple RDF resources
        if self.entity.keyword:
            for keyword in self.entity.keyword.split(u','):
                add_literal_to_graph(dataset_uri, reg.normalize_uri('dcat:keyword'),
                                     keyword.strip(), graph)
        # dct:accrualPeriodicity.
        reg.register_prefix('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
        for concept in self.entity.dataset_frequency:
            frequency_uri = concept.cwuri
            add_uri_to_graph(dataset_uri, reg.normalize_uri('dcterms:accrualPeriodicity'),
                             frequency_uri, graph)
            add_uri_to_graph(frequency_uri, reg.normalize_uri('rdf:type'),
                             reg.normalize_uri('dcterms:Frequency'), graph)
            add_literal_to_graph(frequency_uri, reg.normalize_uri('rdf:value'), concept.label(),
                                 graph)
        # dct:type
        if self.entity.dataset_type:
            dtype = self.entity.dataset_type[0]
            add_concept_to_graph(dtype, dataset_uri, 'dcterms:type', graph, reg)
        # dcat:theme
        for concept in self.entity.dcat_theme:
            add_concept_to_graph(concept, dataset_uri, 'dcat:theme', graph, reg)
        # dct:spatial
        for location in self.entity.spatial_coverage:
            fill_graph_with(location, graph, self.__regid__)
            graph_add_concept_as(graph, reg, location, dataset_uri,
                                 'dcterms:spatial', 'dcterms:Location')


class DistributionRDFAdapter(AbstractRDFAdapter):
    """Adapt Distribution entities to RDF using DCAT vocabulary."""
    __regid__ = 'RDFPrimary'
    __select__ = AbstractRDFAdapter.__select__ & is_instance('Distribution')

    register_rdf_mapping = staticmethod(register_distribution_rdf_mapping)

    def fill(self, graph):
        super(DistributionRDFAdapter, self).fill(graph)
        reg = self.registry
        distribution_uri = rdfio.permanent_url(self.entity)
        # Export access_url and download_url attributes as an RDF resource
        access_url = self.entity.access_url
        if access_url is None:
            downloadable = self.entity.cw_adapt_to('IDownloadable')
            if downloadable:
                # Distribution has a File related.
                access_url = downloadable.download_url()
            else:
                # Fall back to Distribution's URI.
                access_url = rdfio.permanent_url(self.entity)
        add_uri_to_graph(distribution_uri, reg.normalize_uri('dcat:accessURL'),
                         access_url, graph)
        if self.entity.download_url:
            add_uri_to_graph(distribution_uri, reg.normalize_uri('dcat:downloadURL'),
                             self.entity.download_url, graph)
        # Export license from either the distribution directly or from the
        # catalog if distribution has no license.
        license = None
        if self.entity.license:
            license = self.entity.license[0]
        elif self.entity.catalog_license:
            license = self.entity.catalog_license[0]
        if license is not None:
            add_license_to_graph(license, distribution_uri, graph, reg)
        # Export distribution_format relation if available
        for filetype in self.entity.distribution_format:
            fill_graph_with(filetype, graph, self.__regid__)
            graph_add_concept_as(
                graph, reg, filetype, distribution_uri,
                'dcterms:format', 'dcterms:MediaTypeOrExtent')
        # Export media_type if available
        for mediatype in self.entity.media_type:
            fill_graph_with(mediatype, graph, self.__regid__)
            graph_add_concept_as(
                graph, reg, mediatype, distribution_uri,
                'dcat:mediaType', 'dcterms:MediaTypeOrExtent')
        # Documentation
        if self.entity.documentation:
            reg.register_prefix('foaf', 'http://xmlns.com/foaf/0.1/')
            add_uri_to_graph(distribution_uri, reg.normalize_uri('foaf:page'),
                             self.entity.documentation, graph)
        # Export file's metadata.
        if self.entity.reverse_file_distribution:
            distfile = self.entity.reverse_file_distribution[0]
            sha1hex = distfile.data_sha1hex
            if sha1hex is not None:
                add_checksum_to_graph(sha1hex, distribution_uri, graph, reg)


class PeriodOfTimeRDFAdapter(AbstractRDFAdapter):
    """Adapt PeriodOfTime entities to RDF using DCAT vocabulary."""
    __regid__ = 'RDFPrimary'
    __select__ = AbstractRDFAdapter.__select__ & is_instance('PeriodOfTime')
    register_rdf_mapping = staticmethod(register_periodoftime_rdf_mapping)


class AgentRDFAdapter(AbstractRDFAdapter):
    """Adapt Agent entities to foaf:Agent RDF using FOAF vocabulary."""
    __abstract__ = True

    register_rdf_mapping = staticmethod(register_publisher_rdf_mapping)

    def fill(self, graph):
        super(AgentRDFAdapter, self).fill(graph)
        self.fill_email(graph)

    def fill_email(self, graph):
        raise NotImplementedError()


class PublisherRDFAdapter(AgentRDFAdapter):
    """Adapt publisher Agent entities to foaf:Agent RDF using FOAF vocabulary.
    """
    __regid__ = 'RDFPrimary'
    __select__ = (AgentRDFAdapter.__select__ &
                  (has_related_entities('dataset_publisher', 'object') |
                   has_related_entities('catalog_publisher', 'object')))

    register_rdf_mapping = staticmethod(register_publisher_rdf_mapping)

    def fill(self, graph):
        super(PublisherRDFAdapter, self).fill(graph)
        self.fill_publisher_type(graph)

    def fill_publisher_type(self, graph):
        """Export publisher_type relation and delegate to Concept adapter to
        fill object side of the relation.
        """
        if self.entity.publisher_type:
            concept = self.entity.publisher_type[0]
            add_concept_to_graph(concept, rdfio.permanent_url(self.entity),
                                 'dcterms:type', graph, self.registry)

    def fill_email(self, graph):
        """Export email attribute as a owl:Thing RDF resource"""
        reg = self.registry
        reg.register_prefix('foaf', 'http://xmlns.com/foaf/0.1/')
        if self.entity.email:
            add_uri_to_graph(rdfio.permanent_url(self.entity),
                             reg.normalize_uri('foaf:mbox'),
                             u'mailto:{0}'.format(self.entity.email), graph)


class ContactPointRDFAdapter(AgentRDFAdapter):
    """Adapt Agent entities to vcard:Kind RDF using vCard vocabulary."""
    __regid__ = 'RDFContactPoint'
    __select__ = (AgentRDFAdapter.__select__ &
                  has_related_entities('dataset_contact_point', 'object'))

    register_rdf_mapping = staticmethod(register_contact_point_rdf_mapping)

    def fill_email(self, graph):
        reg = self.registry
        # Export email attribute as a owl:Thing RDF resource
        reg.register_prefix('vcard', 'http://www.w3.org/2006/vcard/ns#')
        if self.entity.email:
            add_uri_to_graph(rdfio.permanent_url(self.entity),
                             reg.normalize_uri('vcard:hasEmail'),
                             u'mailto:{0}'.format(self.entity.email), graph)



from cubes.datacat.vocabularies import capitalized_eurovoc_domain

sync_schema_props_perms('language')

# Keep only first letter of Eurovoc concept scheme titles uppercase
domains = cnx.execute('ConceptScheme CS WHERE CS cwuri LIKE "%%eurovoc.europa.eu%%"').entities()
for domain in domains:
    domain.cw_set(title=capitalized_eurovoc_domain(domain.title))
commit(ask_confirm=False)



# copyright 2014-2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-datacat views/forms/actions/components for web ui"""

from functools import partial
from copy import deepcopy
import json
import re
import urlparse

from logilab.mtconverter import (
    xml_escape,
)

from cubicweb import (
    _,
    tags,
)
from cubicweb.predicates import (
    adaptable,
    has_related_entities,
    is_instance,
    match_user_groups,
    one_line_rset,
    score_entity,
)
from cubicweb.uilib import js
from cubicweb.utils import json_dumps
from cubicweb.web import (
    action,
    component,
    facet as cwfacet,
    formwidgets as fw,
    Redirect,
)
from cubicweb.web.views import (
    actions,
    ajaxcontroller,
    baseviews,
    ibreadcrumbs,
    idownloadable,
    json as jsonviews,
    navigation,
    uicfg,
)
from cubicweb.view import (
    AnyRsetView,
    EntityView,
)

from cubes.relationwidget import views as rwdg
from cubes.skos import rdfio
from cubes.skos.views import rdf as rdfviews
from cubes.datacat import opendefinition_licenses
from cubes.datacat.entities import license_rdf_uri


abaa = uicfg.actionbox_appearsin_addmenu
afs = uicfg.autoform_section
affk = uicfg.autoform_field_kwargs
pvdc = uicfg.primaryview_display_ctrl
pvs = uicfg.primaryview_section


uicfg.indexview_etype_section['TransformationSequence'] = 'subobject'
# already shown in "administration" section
uicfg.indexview_etype_section['SKOSSource'] = 'subobject'


def build_selectable_uicfg(rtag, select):
    """Return a copy of uicfg `rtag` with `select` set."""
    copy = deepcopy(rtag)
    copy.__module__ = __name__
    copy.__select__ = select
    return copy


def _autocomplete_widget(fname, wdgclass=fw.LazyRestrictedAutoCompletionWidget,
                         placeholder=None):
    if placeholder is None:
        placeholder = _("Start to type your choice: you'll be presented a list of "
                        "matching terms")
    return wdgclass(
        attrs={'placeholder': placeholder, 'size': 82},
        autocomplete_initfunc=fname,
        autocomplete_settings={'limit': 100, 'delay': 300},
    )


def search_in_scheme(req, scheme_uri):
    """Query ConceptScheme with `scheme_uri` for a Label matching `q` term
    found in request's form.
    """
    term = req.form['q']
    limit = req.form.get('limit', 50)
    lang = req.lang
    qs = ('DISTINCT Any X LIMIT {0} '
          'WHERE X in_scheme S, S cwuri %(scheme)s, L label_of X, L label LL,'
          '      L label ILIKE %(term)s'.format(limit))
    args = {'scheme': scheme_uri, 'term': u'%%%s%%' % term}
    return [{'value': concept.eid, 'label': concept.label(language_code=lang)}
            for concept in req.execute(qs, args).entities()]


class NoRedirectAddRelatedActions(actions.AddRelatedActions):
    """Override 'addrelated' action to avoid redirection to the parent entity."""

    __select__ = (actions.AddRelatedActions.__select__ &
                  is_instance('Agent', 'DataCatalog', 'Dataset', 'Distribution'))

    def redirect_params(self, *args, **kwargs):
        return {}


class WorkflowableEntityOutOfContextView(baseviews.OutOfContextView):
    """Out of context view showing workflow state"""

    __select__ = baseviews.OutOfContextView.__select__ & adaptable('IWorkflowable')

    def cell_call(self, row, col, **kwargs):
        self.w(u'<div style="margin-top: 1px;" class="clearfix">')
        super(WorkflowableEntityOutOfContextView, self).cell_call(row, col, **kwargs)
        entity = self.cw_rset.get_entity(row, col)
        iwf = entity.cw_adapt_to('IWorkflowable')
        self.w(tags.span(iwf.printable_state, klass='badge pull-right'))
        self.w(u'</div>')


#
# RDF views
#

class DatasetDCATRDFView(rdfviews.RDFPrimaryView):
    """RDF view for Dataset entities, outputting complete information about
    catalog, distributions and agents.

    This is mainly useful for clients who *do not follow URIs* to gather information about related
    resources and expect all information at one URL (eg. CKAN).
    """

    __regid__ = 'dcat.rdf.complete'
    __select__ = rdfviews.RDFPrimaryView.__select__ & is_instance('Dataset')
    # (rtype, role, <adapter regid>) of related entities to export along with
    # the dataset.
    follow_related = [
        ('in_catalog', 'subject', 'RDFPrimary'),
        ('of_dataset', 'object', 'RDFPrimary'),
        ('dataset_publisher', 'subject', 'RDFPrimary'),
        ('dataset_contact_point', 'subject', 'RDFContactPoint'),
        ('temporal_coverage', 'subject', 'RDFPrimary'),
        ('locate_datasets', 'object', 'RDFPrimary'),
    ]

    def entity_call(self, entity, graph=None):
        # Copied from parent class
        dump = graph is None
        if graph is None:
            graph = rdfio.default_graph()
        rdfgenerator = entity.cw_adapt_to(self.adapter)
        rdfgenerator.fill(graph)
        # Also output related entities
        for rtype, role, adapter in self.follow_related:
            for related in entity.related(rtype, role, entities=True):
                rdfgenerator = related.cw_adapt_to(adapter)
                rdfgenerator.fill(graph)
        # Copied from parent class
        if dump:
            self._dump(graph)


class DataCatalogDCATRDFView(rdfviews.RDFPrimaryView):
    """RDF view outputting complete information about catalog and children entities.

    This is mainly useful for clients who *do not follow URIs* to gather information about related
    resources and expect all information at one URL (eg. CKAN).
    """

    __regid__ = 'dcat.rdf.complete'
    __select__ = (rdfviews.RDFPrimaryView.__select__ &
                  has_related_entities('in_catalog', role='object'))

    def entity_call(self, entity, graph=None):
        # Copied from parent class
        dump = graph is None
        if graph is None:
            graph = rdfio.default_graph()
        for dataset in entity.related('in_catalog', role='object').entities():
            dataset.view(self.__regid__, graph=graph)
        # Copied from parent class
        if dump:
            self._dump(graph)


class DataCatalogCkanDCATRDFView(DataCatalogDCATRDFView):
    """Backward compatible view calling 'dcat.rdf.complete'."""
    __regid__ = 'dcat.rdf.ckan'


class DatasetCkanDCATRDFView(DatasetDCATRDFView):
    """Backward compatible view calling 'dcat.rdf.complete'."""
    __regid__ = 'dcat.rdf.ckan'


class RDFExportAction(action.Action):

    __regid__ = 'dcat.rdf-export'
    __select__ = one_line_rset() & is_instance('Dataset', 'DataCatalog')
    title = _('RDF export')

    def url(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        return entity.absolute_url(vid='dcat.rdf.complete')


#
# JSON views
#


class CkanJsonLicenseView(jsonviews.JsonMixIn, EntityView):
    """JSON view outputting data about a license that CKAN can use as ``licenses_group_url``. """

    __regid__ = 'dcat.ckan.json.license'
    __select__ = (EntityView.__select__ & score_entity(lambda c: in_license_scheme(c)))

    def entity_call(self, entity, **kwargs):
        od_licenses = opendefinition_licenses()
        url2lic = {}
        title2lic = {}
        for lic_dict in od_licenses.values():
            title2lic[lic_dict['title'].lower()] = lic_dict
            if lic_dict['url']:
                urltuple = urlparse.urlsplit(lic_dict['url'])  # Remove {http://,https://} from URL
                url = urltuple.netloc + urltuple.path
                url2lic[url] = lic_dict
        # Build JSON license
        title = entity.label(language_code=u'en')
        url = license_rdf_uri(entity)
        urltuple = urlparse.urlsplit(url)
        url0 = urltuple.netloc + urltuple.path
        found = url2lic.get(url0, {})  # Find an Open Definition license with same URL
        if not found:
            for title_variant in (  # Find an Open Definition license with same title
                title,
                title.replace('License', 'Licence'),
                re.sub(r'v(\d)', lambda match: match.group(1), title),  # v1.0 -> 1.0
                re.sub(r'v(\d)', lambda match: match.group(1),
                       title.replace('License', 'Licence')),
            ):
                found = title2lic.get(title_variant.lower(), {})
                if found:
                    break
        license = {'title': title, 'url': found.get('url') or url}
        for key, default_value in [
            ('domain_content', False),
            ('domain_data', True),
            ('domain_software', False),
            ('family', u''),
            ('id', entity.absolute_url()),
            ('maintainer', u''),
            ('od_conformance', u'not reviewed'),
            ('osd_conformance', u'not reviewed'),
            ('status', u'active'),
        ]:
            license[key] = found.get(key, default_value)
        self.w(json_dumps(license, sort_keys=True))


class CkanJsonLicensesAction(action.Action):

    __regid__ = 'dcat.ckan.json.licenses'
    __select__ = (is_instance('ConceptScheme')
                  & score_entity(lambda cs: cs.scheme_relation and
                                 cs.scheme_relation[0].name == u'license_type'))
    title = _('JSON export for CKAN')

    def url(self):
        return self._cw.build_url(vid=self.__regid__,
                                  rql=self.cw_rset.printable_rql())


class CkanJsonLicensesView(jsonviews.JsonMixIn, AnyRsetView):
    """JSON view outputting data about a license that CKAN can use as ``licenses_group_url``. """

    __regid__ = 'dcat.ckan.json.licenses'
    __select__ = (AnyRsetView.__select__ &
                  score_entity(lambda cs: cs.scheme_relation and
                               cs.scheme_relation[0].name == u'license_type'))

    def call(self, **kwargs):
        licenses = []
        for scheme in self.cw_rset.entities():
            for concept in scheme.reverse_in_scheme:
                licenses.append(
                    json.loads(concept.view('dcat.ckan.json.license'))
                )
        self.w(json_dumps(licenses, sort_keys=True))


#
# HTML views
#

# DCAT-AP restrictions.
# Set some relations as "required" to conform to DCAT-AP recommandations
# despite our schema matches plain DCAT which is more permissive. Only DCAT-AP
# profile is supported through the UI at the moment.
affk.tag_attribute(('DataCatalog', 'title'), {'required': True})
affk.tag_attribute(('DataCatalog', 'description'), {'required': True})
affk.tag_attribute(('Dataset', 'title'), {'required': True})
affk.tag_attribute(('Dataset', 'description'), {'required': True})
affk.tag_subject_of(('DataCatalog', 'catalog_publisher', '*'),
                    {'required': True})


class NoPreviewIDownloadablePrimaryView(idownloadable.IDownloadablePrimaryView):
    """IDownloadablePrimaryView without data preview."""

    def render_data(self, *args):
        return False


class IDownloadableOutOfContextView(baseviews.OutOfContextView):
    """Out of context view for IDownloadable

    Adapted from cubicweb.web.views.idownloadable.IDownloadableOneLineView
    with content-type and a download icon.

    This is used in `file_distribution` relation contextual component in
    particular.
    """

    __select__ = (baseviews.OutOfContextView.__select__ &
                  adaptable('IDownloadable'))
    download_icon = 'glyphicon glyphicon-download-alt'

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        url = xml_escape(entity.absolute_url())
        adapter = entity.cw_adapt_to('IDownloadable')
        name = xml_escape(entity.dc_title())
        durl = xml_escape(adapter.download_url())
        dcontenttype = xml_escape(adapter.download_content_type())
        self.w(u'<a class="{2}" href="{0}" title="{1}"></a> '.format(
               durl, self._cw._('download'), self.download_icon))
        self.w(u'<a href="{0}">{1}</a> '.format(url, name))
        self.w(u'<span class="badge">{0}</span>'.format(dcontenttype))


# File
uicfg.indexview_etype_section['File'] = 'subobject'
pvs.tag_subject_of(('File', 'replaces', '*'), 'hidden')
pvs.tag_object_of(('*', 'replaces', 'File'), 'hidden')
afs.tag_object_of(('*', 'replaces', 'File'), 'main', 'hidden')
abaa.tag_object_of(('*', 'replaces', 'File'), True)

# afs.tag_subject_of(('File', 'resource_of', '*'), 'main', 'hidden')
afs.tag_object_of(('*', 'process_input_file', 'File'), 'main', 'hidden')

pvs.tag_subject_of(('File', 'file_distribution', '*'), 'hidden')


navigation.NextPrevNavigationComponent.context = 'navtop'


class FileIPrevNextAdapter(navigation.IPrevNextAdapter):
    """IPrevNext adapter for file replaced by or replacing another file.
    """

    __select__ = (has_related_entities('replaces', role='subject') |
                  has_related_entities('replaces', role='object'))

    def next_entity(self):
        successors = self.entity.reverse_replaces
        if successors:
            return successors[0]

    def previous_entity(self):
        predecessors = self.entity.replaces
        if predecessors:
            return predecessors[0]


class FileReplacedBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    """Define <New file version>/<Old file version> breadcrumb."""

    __select__ = (ibreadcrumbs.IBreadCrumbsAdapter.__select__ &
                  has_related_entities('replaces', role='object'))

    def parent_entity(self):
        return self.entity.reverse_replaces[0]


class DistributionFileBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    """Define Distribution / File breadcrumbs"""
    __select__ = (ibreadcrumbs.IBreadCrumbsAdapter.__select__ &
                  has_related_entities('file_distribution', role='subject') &
                  # Prevent select ambiguity ad File can be object of both
                  # `implemented_by` and `file_distribution` relations.
                  ~has_related_entities('implemented_by', role='object'))

    def parent_entity(self):
        return self.entity.file_distribution[0]


# Agent
abaa.tag_object_of(('*', 'catalog_publisher', 'Agent'), True)
afs.tag_object_of(('*', 'catalog_publisher', 'Agent'), 'main', 'hidden')
affk.set_field_kwargs('Agent', 'name', widget=fw.TextInput({'size': 82}))
affk.set_field_kwargs('Agent', 'email', widget=fw.TextInput({'size': 82}))
pvs.tag_subject_of(('Agent', 'publisher_type', 'Concept'), 'attributes')
afs.tag_subject_of(('Agent', 'publisher_type', 'Concept'), 'main', 'attributes')


# Concept

pvs.tag_object_of(('*', 'catalog_license', 'Concept'), 'hidden')


def scheme_relation_choices(form, field, **kwargs):
    """Choices function for CheckBox widget to select/unselect whether a scheme
    is used to qualify a license_type relationship.
    """
    choices = [(form._cw._('no specific usage'), u'')]
    for name in (u'license_type', u'dataset_type'):
        rtype_eid = form._cw.find('CWRType', name=name)[0][0]
        choices.append((form._cw._(name), unicode(rtype_eid)))
    return choices


def scheme_relation_value(form, field, **kwargs):
    """Return current value of scheme_relation."""
    edited = form.edited_entity
    if edited.scheme_relation:
        rtype = edited.scheme_relation[0]
        return (rtype.name, rtype.eid)
    return u''


def in_license_scheme(entity):
    """Return 1 if entity is a Concept in a scheme used to describe license
    type.
    """
    if not entity.cw_etype == 'Concept':
        return 0
    if not entity.has_eid():
        return 0
    scheme = entity.in_scheme[0]
    if (scheme.scheme_relation and
            scheme.scheme_relation[0].name == u'license_type'):
        return 1
    return 0


pvs.tag_subject_of(('Concept', 'license_type', 'Concept'), 'hidden')

license_concept_pvs = build_selectable_uicfg(pvs, score_entity(in_license_scheme))
license_concept_pvs.tag_subject_of(('Concept', 'license_type', 'Concept'), 'attributes')


afs.tag_subject_of(('Concept', 'license_type', 'Concept'), 'main', 'hidden')
afs.tag_object_of(('Concept', 'license_type', 'Concept'), 'main', 'hidden')
for rtype in ('publisher_type', 'language', 'dataset_frequency', 'dcat_theme',
              'license', 'media_type', 'distribution_format'):
    afs.tag_object_of(('*', rtype, 'Concept'), 'main', 'hidden')
afs.tag_subject_of(('Concept', 'license', '*'), 'main', 'hidden')

licence_concept_afs = build_selectable_uicfg(afs, score_entity(in_license_scheme))
licence_concept_afs.tag_subject_of(('Concept', 'license_type', 'Concept'), 'main', 'attributes')
afs.tag_object_of(('*', 'theme_taxonomy', 'ConceptScheme'), 'main', 'hidden')
afs.tag_subject_of(('ConceptScheme', 'scheme_relation', '*'), 'main', 'attributes')
affk.tag_subject_of(('ConceptScheme', 'scheme_relation', '*'),
                    {'label': _('usage'),
                     'help': _('usage of concepts in this scheme'),
                     'widget': fw.Radio,
                     'value': scheme_relation_value,
                     'fallback_on_none_attribute': True,
                     'sort': True,
                     'choices': scheme_relation_choices})


# DataCatalog
pvs.tag_attribute(('DataCatalog', 'title'), 'hidden')
pvs.tag_attribute(('DataCatalog', 'issued'), 'hidden')
afs.tag_attribute(('DataCatalog', 'issued'), 'main', 'hidden')
pvs.tag_attribute(('DataCatalog', 'modified'), 'hidden')
afs.tag_attribute(('DataCatalog', 'modified'), 'main', 'hidden')
pvs.tag_subject_of(('DataCatalog', 'catalog_publisher', '*'), 'hidden')
pvs.tag_subject_of(('DataCatalog', 'license', '*'), 'attributes')
pvdc.tag_subject_of(('DataCatalog', 'license', '*'), {'vid': 'incontext'})
afs.tag_subject_of(('DataCatalog', 'theme_taxonomy', '*'), 'main', 'attributes')
afs.tag_subject_of(('DataCatalog', 'catalog_publisher', '*'), 'main', 'attributes')
abaa.tag_object_of(('*', 'in_catalog', 'DataCatalog'), True)
abaa.tag_subject_of(('DataCatalog', 'theme_taxonomy', '*'), True)
afs.tag_object_of(('*', 'in_catalog', 'DataCatalog'), 'main', 'hidden')
pvdc.tag_attribute(('DataCatalog', 'homepage'), {'vid': 'urlattr'})
for attr in ('title', 'homepage'):
    affk.set_field_kwargs('DataCatalog', attr, widget=fw.TextInput({'size': 82}))
afs.tag_subject_of(('DataCatalog', 'license', 'Concept'), 'main', 'attributes')
pvs.tag_subject_of(('DataCatalog', 'language', '*'), 'attributes')
afs.tag_subject_of(('DataCatalog', 'language', '*'), 'main', 'attributes')
affk.tag_subject_of(('DataCatalog', 'language', '*'),
                    {'widget': rwdg.RelationFacetWidget(no_creation_form=True)})


class DataCatalogBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    """Define <Agent>/<DataCatalog> breadcrumb."""

    __select__ = (ibreadcrumbs.IBreadCrumbsAdapter.__select__ &
                  has_related_entities('catalog_publisher'))

    def parent_entity(self):
        return self.entity.catalog_publisher[0]


pvs.tag_subject_of(('*', 'spatial_coverage', '*'), 'attributes')
afs.tag_subject_of(('*', 'spatial_coverage', '*'), 'main', 'attributes')
affk.tag_subject_of(('*', 'spatial_coverage', '*'),
                    {'widget': rwdg.RelationFacetWidget(no_creation_form=True)})


# Dataset

def dataset_publisher_choices(form, field, **kwargs):
    """Return form field choices corresponding to all possible publishers for a dataset."""
    # Try to find out if this dataset is linked to a catalog
    linkedto = form.linked_to.get(('in_catalog', 'subject'))
    if linkedto:  # Adding dataset from a catalog view
        catalog = form._cw.entity_from_eid(linkedto[0])
        publishers = catalog.catalog_publisher
        publisher = publishers[0] if publishers else None
        if publisher is not None:
            return [(publisher.dc_title(), unicode(publisher.eid))]
    limit = kwargs.get('limit', 50)
    qs = 'DISTINCT Agent X LIMIT {limit}'.format(limit=limit)
    return [(entity.dc_title(), unicode(entity.eid)) for entity in form._cw.execute(qs).entities()]


pvs.tag_subject_of(('Dataset', 'in_catalog', '*'), 'hidden')
pvs.tag_subject_of(('Dataset', 'dataset_frequency', '*'), 'attributes')
pvs.tag_subject_of(('Dataset', 'dataset_contact_point', '*'), 'attributes')
pvs.tag_subject_of(('Dataset', 'dataset_publisher', '*'), 'attributes')
pvs.tag_subject_of(('Dataset', 'temporal_coverage', '*'), 'attributes')
pvdc.tag_attribute(('Dataset', 'landing_page'), {'vid': 'urlattr'})
afs.tag_subject_of(('Dataset', 'dataset_contact_point', '*'), 'main', 'attributes')
afs.tag_subject_of(('Dataset', 'dataset_publisher', '*'), 'main', 'attributes')
pvs.tag_subject_of(('Dataset', 'dataset_contributors', '*'), 'attributes')
afs.tag_subject_of(('Dataset', 'dataset_contributors', '*'), 'main', 'attributes')
afs.tag_subject_of(('Dataset', 'temporal_coverage', '*'), 'main', 'inlined')
abaa.tag_subject_of(('Dataset', 'temporal_coverage', '*'), False)
pvs.tag_attribute(('Dataset', 'title'), 'hidden')
pvs.tag_attribute(('Dataset', 'identifier'), 'hidden')
afs.tag_attribute(('Dataset', 'identifier'), 'main', 'hidden')
pvs.tag_attribute(('Dataset', 'issued'), 'hidden')
afs.tag_attribute(('Dataset', 'issued'), 'main', 'hidden')
pvs.tag_attribute(('Dataset', 'modified'), 'hidden')
afs.tag_attribute(('Dataset', 'modified'), 'main', 'hidden')
# Keep dcat_theme hidden from autoform since the RelationWidget will not work
# on creation as constraints on DataCatalog cannot be checked and we prefer
# the user to edit this *inline* from the primary view.
afs.tag_subject_of(('Dataset', 'dcat_theme', '*'), 'main', 'hidden')
pvs.tag_subject_of(('Dataset', 'dcat_theme', '*'), 'attributes')
affk.tag_subject_of(('Dataset', 'dcat_theme', '*'),
                    {'widget': rwdg.RelationFacetWidget(no_creation_form=True)})
pvs.tag_subject_of(('Dataset', 'dataset_type', '*'), 'attributes')
afs.tag_subject_of(('Dataset', 'dataset_type', '*'), 'main', 'attributes')
affk.tag_subject_of(('Dataset', 'dataset_type', '*'),
                    {'widget': rwdg.RelationFacetWidget(no_creation_form=True)})
afs.tag_subject_of(('Dataset', 'in_catalog', '*'), 'main', 'attributes')
afs.tag_object_of(('*', 'resource_feed_of', 'Dataset'), 'main', 'hidden')
afs.tag_object_of(('*', 'of_dataset', 'Dataset'), 'main', 'hidden')
for attr in ('identifier', 'title', 'keyword', 'landing_page',
             'provenance'):
    affk.set_field_kwargs('Dataset', attr, widget=fw.TextInput({'size': 82}))
afs.tag_subject_of(('Dataset', 'dataset_frequency', 'Concept'),
                   'main', 'attributes')
affk.tag_subject_of(('Dataset', 'dataset_publisher', '*'),
                    {'choices': dataset_publisher_choices})

pvs.tag_object_of(('*', 'locate_datasets', 'Dataset'), 'hidden')
afs.tag_object_of(('*', 'locate_datasets', 'Dataset'), 'main', 'attributes')
affk.tag_object_of(('*', 'locate_datasets', 'Dataset'),
                   {'widget': rwdg.RelationFacetWidget()})


class DatasetPublisherFacet(cwfacet.RelationFacet):
    __regid__ = 'datacat.dataset_publisher_facet'
    __select__ = cwfacet.RelationFacet.__select__ & is_instance('Dataset')
    rtype = 'dataset_publisher'
    role = 'subject'
    target_attr = 'name'


class DatasetInCatalogFacet(cwfacet.RelationFacet):
    __regid__ = 'datacat.dataset_in_catalog_facet'
    __select__ = cwfacet.RelationFacet.__select__ & is_instance('Dataset')
    rtype = 'in_catalog'
    role = 'subject'
    target_attr = 'title'


class DatasetIssuedFacet(cwfacet.DateRangeFacet):
    """Facet filtering datasets based on their publication date."""
    __regid__ = 'datacat.dataset_issued_facet'
    __select__ = cwfacet.DateRangeFacet.__select__ & is_instance('Dataset')
    rtype = 'issued'


class DatasetModifiedFacet(cwfacet.DateRangeFacet):
    """Facet filtering datasets based on their modification date."""
    __regid__ = 'datacat.dataset_modified_facet'
    __select__ = cwfacet.DateRangeFacet.__select__ & is_instance('Dataset')
    rtype = 'modified'


class DatasetCatalogBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    """Define DataCatalog / Dataset breadcrumbs"""
    __select__ = (ibreadcrumbs.IBreadCrumbsAdapter.__select__ &
                  has_related_entities('in_catalog', role='subject'))

    def parent_entity(self):
        return self.entity.in_catalog[0]


class DatasetHasDistributionFacet(cwfacet.HasRelationFacet):
    """Facet filtering datasets with a distribution."""
    __regid__ = 'datacat.has_distribution'
    __select__ = cwfacet.HasRelationFacet.__select__ & is_instance('Dataset')
    rtype, role = 'of_dataset', 'object'


class ConceptSchemeUsageFacet(cwfacet.RelationFacet):
    __regid__ = 'datacat.conceptschema_usage_facet'
    __select__ = cwfacet.RelationFacet.__select__ & is_instance('ConceptScheme')
    rtype = 'scheme_relation'
    role = 'subject'
    target_attr = 'name'
    title = _('usage')


# Entity types with a media_type relation

@ajaxcontroller.ajaxfunc(output_type='json')
def get_mediattypes(self):
    return search_in_scheme(
        self._cw, 'http://www.iana.org/assignments/media-types/media-types.xml')


affk.tag_subject_of(
    ('*', 'media_type', 'Concept'),
    {'widget': _autocomplete_widget(
        'get_mediattypes',
        placeholder=_('e.g. text/csv, image/gif, application/octet-stream, etc.'))
     },
)


# Distribution

class DistributionBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    """Define Dataset / Distribution breadcrumbs"""
    __select__ = (ibreadcrumbs.IBreadCrumbsAdapter.__select__ &
                  has_related_entities('of_dataset', role='subject'))

    def parent_entity(self):
        return self.entity.of_dataset[0]


pvs.tag_attribute(('Distribution', 'title'), 'hidden')
affk.set_field_kwargs('Distribution', 'title', widget=fw.TextInput({'size': 82}))
afs.tag_object_of(('*', 'file_distribution', 'Distribution'), 'main', 'inlined')
abaa.tag_object_of(('*', 'file_distribution', 'Distribution'), True)
afs.tag_attribute(('Distribution', 'byte_size'), 'main', 'hidden')
afs.tag_attribute(('Distribution', 'download_url'), 'main', 'hidden')
afs.tag_attribute(('Distribution', 'issued'), 'main', 'hidden')
afs.tag_attribute(('Distribution', 'modified'), 'main', 'hidden')
pvs.tag_subject_of(('Distribution', 'license', '*'), 'attributes')
pvdc.tag_subject_of(('Distribution', 'license', '*'), {'vid': 'incontext'})
pvs.tag_subject_of(('Distribution', 'media_type', '*'), 'attributes')
pvs.tag_subject_of(('Distribution', 'distribution_format', '*'), 'attributes')
afs.tag_subject_of(('Distribution', 'license', '*'), 'main', 'inlined')
affk.tag_subject_of(
    ('Distribution', 'distribution_format', 'Concept'),
    {'widget': _autocomplete_widget(
        'get_filetypes', placeholder=_('e.g. CSV, PNG, etc.'))}
)
afs.tag_subject_of(('Distribution', 'license', 'Concept'), 'main', 'attributes')


@ajaxcontroller.ajaxfunc(output_type='json')
def get_filetypes(self):
    return search_in_scheme(
        self._cw, 'http://publications.europa.eu/resource/authority/file-type')


pvs.tag_attribute(('Distribution', 'access_url'), 'hidden')
afs.tag_attribute(('Distribution', 'access_url'), 'main', 'hidden')
pvs.tag_attribute(('Distribution', 'download_url'), 'hidden')
afs.tag_attribute(('Distribution', 'download_url'), 'main', 'hidden')
pvs.tag_subject_of(('Distribution', 'of_dataset', '*'), 'hidden')
afs.tag_subject_of(('Distribution', 'of_dataset', '*'), 'main', 'hidden')
pvs.tag_subject_of(('Distribution', 'distribution_format', '*'), 'attributes')
afs.tag_subject_of(('Distribution', 'distribution_format', '*'), 'main', 'attributes')
pvs.tag_subject_of(('Distribution', 'media_type', '*'), 'attributes')
afs.tag_subject_of(('Distribution', 'media_type', '*'), 'main', 'attributes')
pvs.tag_attribute(('Distribution', 'issued'), 'hidden')
afs.tag_attribute(('Distribution', 'issued'), 'main', 'hidden')
pvs.tag_attribute(('Distribution', 'modified'), 'hidden')
afs.tag_attribute(('Distribution', 'modified'), 'main', 'hidden')
pvs.tag_attribute(('Distribution', 'byte_size'), 'hidden')
afs.tag_attribute(('Distribution', 'byte_size'), 'main', 'hidden')

# XXX does not work except in creation mode.
distr_afs = build_selectable_uicfg(afs, has_related_entities('file_distribution'))
distr_afs.tag_attribute(('File', 'title'), 'main', 'hidden')
distr_afs.tag_attribute(('File', 'description'), 'main', 'hidden')

# ResourceFeed
pvdc.tag_attribute(('ResourceFeed', 'url'), {'vid': 'urlattr'})
pvs.tag_attribute(('ResourceFeed', 'title'), 'hidden')
pvs.tag_subject_of(('ResourceFeed', 'resource_feed_of', '*'), 'hidden')
affk.set_field_kwargs('ResourceFeed', 'title', widget=fw.TextInput({'size': 82}))
affk.set_field_kwargs('ResourceFeed', 'url', widget=fw.TextInput({'size': 82}))
afs.tag_subject_of(('ResourceFeed', 'resource_feed_of', '*'), 'main', 'hidden')
pvs.tag_subject_of(('ResourceFeed', 'media_type', '*'), 'attributes')
afs.tag_subject_of(('ResourceFeed', 'media_type', '*'), 'main', 'attributes')
# No reledit for `media_type` relation since this cannot handle coupling with
# script form field.
pvdc.tag_subject_of(('ResourceFeed', 'media_type', '*'), {'vid': 'incontext'})
affk.set_fields_order('ResourceFeed',
                      ('title', 'url', 'media_type', 'validation_script',
                       'transformation_script'))
pvdc.tag_subject_of(('ResourceFeed', 'media_type', '*'), {'order': 2})
pvdc.tag_subject_of(('ResourceFeed', 'validation_script', '*'), {'order': 3})
pvdc.tag_subject_of(('ResourceFeed', 'transformation_script', '*'), {'order': 4})
pvs.tag_object_of(('*', 'process_for_resourcefeed', 'ResourceFeed'), 'hidden')
afs.tag_object_of(('*', 'process_for_resourcefeed', 'ResourceFeed'), 'main', 'hidden')
afs.tag_subject_of(('ResourceFeed', 'resource_feed_source', '*'),
                   'main', 'hidden')
abaa.tag_object_of(('*', 'process_for_resourcefeed', 'ResourceFeed'), False)


class ResourceFeedMediaTypeMasterWidget(fw.LazyRestrictedAutoCompletionWidget):
    """Custom widget that controls the content of "dependent" widgets."""
    needs_js = fw.LazyRestrictedAutoCompletionWidget.needs_js + ('cubes.datacat.js', )

    def render(self, form, field, renderer=None):
        master_domid = field.dom_id(form)
        dependent_info = dict(
            (etype, form.field_by_name(rtype, 'subject').dom_id(form))
            for etype, rtype in [('ValidationScript', 'validation_script'),
                                 ('TransformationScript', 'transformation_script')]
        )
        form._cw.add_onload(js.cw.cubes.datacat.initDependentFormField(
            master_domid, dependent_info))
        return super(ResourceFeedMediaTypeMasterWidget, self).render(form, field, renderer)


def script_choice(etype, form, field):
    """Vocabulary function grouping entities of type `etype` choices by
    media_type.
    """
    req = form._cw
    rset = req.execute('Any X,N,C,LL WHERE X is {0}, X name N, X media_type C,'
                       ' C preferred_label L, L label LL'.format(etype))
    vdict = {}
    for script_eid, script_name, mtype_eid, mtype in rset:
        vdict.setdefault((mtype_eid, mtype), []).append(
            (script_name, str(script_eid)))
    vocab = [(u'', '')]
    for (mtype_eid, mtype), values in vdict.iteritems():
        optgroup_id = '{0}_mediatype_{1}'.format(
            etype.lower(), mtype.replace('/', '-'))
        vocab.append((mtype, None, {'id': optgroup_id}))
        vocab.extend(values)
    return vocab


for etype, rtype in [('ValidationScript', 'validation_script'),
                     ('TransformationScript', 'transformation_script')]:
    afs.tag_subject_of(('ResourceFeed', rtype, '*'), 'main', 'attributes')
    affk.tag_subject_of(('ResourceFeed', rtype, '*'),
                        {'choices': partial(script_choice, etype),
                         'sort': False})


class CWSourceBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    """Define ResourceFeed / CWSource breadcrumbs"""
    __select__ = (ibreadcrumbs.IBreadCrumbsAdapter.__select__ &
                  has_related_entities('resource_feed_source', role='object'))

    def parent_entity(self):
        return self.entity.reverse_resource_feed_source[0]


class ResourceFeedBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    """Define Dataset / ResourceFeed breadcrumbs"""
    __select__ = (ibreadcrumbs.IBreadCrumbsAdapter.__select__ &
                  has_related_entities('resource_feed_of', role='subject'))

    def parent_entity(self):
        """The Dataset"""
        return self.entity.resource_feed_of[0]


class DataProcessInResourceFeedCtxComponent(component.EntityCtxComponent):
    """Display data processes in ResourceFeed primary view"""
    __regid__ = 'datacat.resourcefeed-dataprocess'
    __select__ = (component.EntityCtxComponent.__select__ &
                  is_instance('ResourceFeed') &
                  has_related_entities('process_for_resourcefeed',
                                       role='object'))
    title = _('Data processes')
    context = 'navcontentbottom'

    def render_body(self, w):
        rset = self._cw.execute(
            'Any P,CD,S,I,E,D ORDERBY CD DESC LIMIT 10 WHERE'
            ' P process_for_resourcefeed X,'
            ' P process_input_file I,'
            ' P in_state S,'
            ' D? process_depends_on P,'
            ' P creation_date CD,'
            ' P process_stderr E?,'
            ' X eid %(eid)s',
            {'eid': self.entity.eid})
        if rset:
            w(self._cw.view('table', rset=rset))
        rset = self._cw.execute(
            'Any P,CD,S,I,E,D ORDERBY CD DESC LIMIT 10 WHERE'
            ' P process_for_resourcefeed X,'
            ' P process_input_file I,'
            ' P in_state S,'
            ' P process_depends_on D?,'
            ' P creation_date CD,'
            ' P process_stderr E?,'
            ' X eid %(eid)s',
            {'eid': self.entity.eid})
        if rset:
            w(self._cw.view('table', rset=rset))


class ResourceFeedCleanProcessesAction(action.Action):
    __regid__ = 'datacat.clean-resourcefeed-process-logs'
    __select__ = (action.Action.__select__ & match_user_groups('managers')
                  & one_line_rset()
                  & has_related_entities('process_for_resourcefeed', role='object'))

    title = _('cleanup process logs')
    category = 'moreactions'

    def url(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        return entity.absolute_url(vid=self.__regid__)


class ResourceFeedCleanProcessesView(EntityView):
    __regid__ = 'datacat.clean-resourcefeed-process-logs'
    __select__ = (EntityView.__select__ & match_user_groups('managers')
                  & one_line_rset()
                  & has_related_entities('process_for_resourcefeed', role='object'))

    def entity_call(self, entity):
        deleted = self._cw.call_service(self.__regid__, eid=entity.eid)
        msg = self._cw._('Number of log files is below specified limit')
        if deleted:
            if deleted == 1:
                msg = self._cw._('Most ancient process log file deleted')
            else:
                msg = self._cw._('%d most ancient process log files deleted'
                                 % len(deleted))
        raise Redirect(entity.absolute_url(__message=msg))


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__,
                      [NoPreviewIDownloadablePrimaryView])
    vreg.register_and_replace(NoPreviewIDownloadablePrimaryView,
                              idownloadable.IDownloadablePrimaryView)



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-datacat tests for views"""

from datetime import datetime
import json
from tempfile import NamedTemporaryFile

from mock import patch
from pytz import utc
from rdflib.compare import (
    graph_diff,
    to_isomorphic,
)

from cubicweb import Binary
from cubicweb.devtools import (
    PostgresApptestConfiguration,
    testlib,
    startpgcluster,
    stoppgcluster,
)

from cubes.skos import rdfio
from cubes.datacat.views import dataset_publisher_choices


def setUpModule():
    startpgcluster(__file__)


def tearDownModule():
    stoppgcluster(__file__)


class PublisherChoicesFunctionTC(testlib.CubicWebTC):
    """Test case for cubicweb-datacat dataset publisher choices function."""

    configcls = PostgresApptestConfiguration

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            pub1 = cnx.create_entity('Agent', name=u'Publisher 1')
            pub2 = cnx.create_entity('Agent', name=u'Publisher 2')
            cat = cnx.create_entity('DataCatalog', title=u'My Catalog',
                                    description=u'A nice catalog', catalog_publisher=pub1)
            cnx.commit()
        self.cat_eid = cat.eid
        self.pub1_eid = pub1.eid
        self.pub2_eid = pub2.eid

    def test_choices_add_dataset_from_catalog(self):
        """Check that publisher list is correct when dataset is created from a catalog."""
        with self.admin_access.web_request() as req:
            req.form['__linkto'] = 'in_catalog:{0}:subject'.format(self.cat_eid)
            dset_class = self.vreg['etypes'].etype_class('Dataset')(req)
            add_form = self.vreg['forms'].select('edition', req, entity=dset_class)
            publisher_field = add_form.field_by_name('dataset_publisher', 'subject')
            self.assertEqual(dataset_publisher_choices(add_form, publisher_field),
                             [(u'Publisher 1', unicode(self.pub1_eid))])

    def test_choices_add_dataset_no_catalog(self):
        """Check that publisher list is correct when dataset is created standalone."""
        with self.admin_access.web_request() as req:
            dset_class = self.vreg['etypes'].etype_class('Dataset')(req)
            add_form = self.vreg['forms'].select('edition', req, entity=dset_class)
            publisher_field = add_form.field_by_name('dataset_publisher', 'subject')
            self.assertEqual(dataset_publisher_choices(add_form, publisher_field),
                             [(u'Publisher 1', unicode(self.pub1_eid)),
                              (u'Publisher 2', unicode(self.pub2_eid))])


def sorted_triples(graph):
    return '\n'.join(sorted(graph.serialize(format='nt').splitlines()))


def conceptscheme(cnx, uri, **kwargs):
    return cnx.create_entity('ConceptScheme', cwuri=uri, **kwargs)


class RDFAdapterTC(testlib.CubicWebTC):
    """Test case for RDF data export."""

    configcls = PostgresApptestConfiguration

    def setup_database(self):
        # patch now() so that computed datetime attributes always return the same datetime
        # See https://docs.python.org/3/library/unittest.mock-examples.html#partial-mocking
        with patch('cubes.datacat.hooks.datetime') as mock_dt:
            mock_dt.now.return_value = datetime(2016, 02, 02, 15, 25, 0, tzinfo=utc)
            mock_dt.side_effect = datetime
            # Now set up entities
            with self.admin_access.repo_cnx() as cnx:
                spatial_scheme = conceptscheme(
                    cnx, u'http://publications.europa.eu/resource/authority/country')
                lic_scheme = conceptscheme(
                    cnx, u'http://publications.europa.eu/resource/authority/licence')
                cnx.execute('SET CS scheme_relation RT WHERE CS eid %(cs)s, RT name %(rt)s',
                            {'cs': lic_scheme.eid, 'rt': 'license_type'})
                scheme = conceptscheme(cnx, u'http://example.org/scheme',
                                       title=u'Concept Scheme')
                nat_concept = scheme.add_concept(u'National authority',
                                                 cwuri=u'http://example.org/national_authority')
                attribution_concept = scheme.add_concept(u'Attribution',
                                                         cwuri=u'http://example.org/attribution')
                annual_concept = scheme.add_concept(u'Annual', cwuri=u'http://example.org/annual')
                csv_concept = scheme.add_concept(u'CSV', cwuri=u'http://example.org/csv')
                xls_concept = scheme.add_concept(u'Excel XLS', cwuri=u'http://example.org/xls')
                appxls_concept = scheme.add_concept(u'application/vnd.ms-excel',
                                                    cwuri=u'http://example.org/application_xls')
                zip_concept = scheme.add_concept(u'ZIP', cwuri=u'http://example.org/zip')
                appzip_concept = scheme.add_concept(u'application/zip',
                                                    cwuri=u'http://example.org/application_zip')
                eng_concept = scheme.add_concept(u'English', cwuri=u'http://example.org/english')
                edu_concept = scheme.add_concept(u'Education, culture and sport',
                                                 cwuri=u'http://example.org/education')
                publisher = cnx.create_entity('Agent', name=u'The Publisher',
                                              publisher_type=nat_concept,
                                              email=u'publisher@example.org',
                                              cwuri=u'http://example.org/publisher')
                contact = cnx.create_entity('Agent', name=u'The Contact Point',
                                            email=u'contact@example.org',
                                            cwuri=u'http://example.org/contact')
                license = cnx.create_entity('Concept', in_scheme=lic_scheme,
                                            license_type=attribution_concept,
                                            cwuri=u'http://example.org/license')
                cnx.create_entity('Label', label_of=license, kind=u"preferred",
                                  label=u'Other (attribution)')
                toulouse = spatial_scheme.add_concept(
                    u'Toulouse', cwuri=u'http://example.org/toulouse')
                parcducanal = spatial_scheme.add_concept(
                    u'Parc tech du canal', cwuri=u'http://example.org/parc-du-canal')
                floor1 = spatial_scheme.add_concept(
                    u'floor 1', cwuri=u'http://example.org/floor1')
                cat = cnx.create_entity('DataCatalog', title=u'My Catalog',
                                        description=u'A nice catalog', catalog_publisher=publisher,
                                        homepage=u'http://cat.example.org', language=eng_concept,
                                        theme_taxonomy=scheme, license=license,
                                        issued=datetime(2016, 02, 01, 20, 40, 0, tzinfo=utc),
                                        modified=datetime(2016, 02, 02, 18, 25, 0, tzinfo=utc),
                                        spatial_coverage=[toulouse, parcducanal],
                                        cwuri=u'http://example.org/catalog')
                dset_type = scheme.add_concept(u'blah', cwuri=u'http://example.org/blah')
                ds = cnx.create_entity('Dataset', title=u'First Dataset', description=u'A dataset',
                                       identifier=u'dataset-1',
                                       landing_page=u'http://www.example.org/dataset-1',
                                       in_catalog=cat, dataset_publisher=publisher,
                                       dataset_contact_point=contact, keyword=u'keyword',
                                       dataset_frequency=annual_concept, dcat_theme=edu_concept,
                                       spatial_coverage=floor1,
                                       dataset_type=dset_type,
                                       cwuri=u'http://example.org/dataset')
                cnx.create_entity(
                    'Location', name=u'somewhere',
                    cwuri=u'http://example.org/location/somewhere',
                    geometry=(u'GEOMETRYCOLLECTION (POINT (2 46))', 4326),
                    locate_datasets=ds)
                cnx.create_entity('PeriodOfTime',
                                  cwuri=u'http://example.org/2012-2013',
                                  start=datetime(2012, 1, 1, tzinfo=utc),
                                  end=datetime(2013, 2, 1, tzinfo=utc),
                                  reverse_temporal_coverage=ds)
                csv_dist = cnx.create_entity(
                    'Distribution', title=u'First Dataset (CSV)',
                    description=u'First Dataset in CSV format',
                    of_dataset=ds, license=license,
                    distribution_format=csv_concept,
                    cwuri=u'http://example.org/dataset/csv')

                with cnx.security_enabled(write=False):
                    # by-pass security on data_sha1hex
                    cnx.create_entity('File', data=Binary(b'a,b,c'),
                                      data_name=u'csv',
                                      data_sha1hex=u'chat1',
                                      file_distribution=csv_dist)

                cnx.create_entity('Distribution', title=u'First Dataset (XLS)',
                                  description=u'First Dataset in XLS format',
                                  of_dataset=ds, license=license,
                                  distribution_format=xls_concept,
                                  media_type=appxls_concept,
                                  access_url=u'http://www.example.org',
                                  cwuri=u'http://example.org/dataset/xls')
                cnx.create_entity('Distribution', title=u'First Dataset (ZIP)',
                                  description=u'First Dataset in ZIP format',
                                  of_dataset=ds,
                                  distribution_format=zip_concept,
                                  media_type=appzip_concept,
                                  documentation=u'http://d.a.ta',
                                  access_url=u'http://www.example.org',
                                  cwuri=u'http://example.org/dataset/zip')
                cnx.commit()
                self.cat_eid = cat.eid

    @patch('cubes.datacat.entities.DownloadableDistribution.download_url',
           return_value=u'http://example.org/abc/download')  # predictable Distribution's access_url
    @patch('cubes.skos.rdfio.permanent_url',  # predicable URIs (no eid)
           side_effect=lambda entity: entity.cwuri)
    @patch('cubes.skos.rdfio.RDFGraphGenerator.same_as_uris')
    def _check_rdf_view(self, action_regid, expected_fname, mock_same_as_uris,
                        mock_permanent_url, mock_download_url):
        # Because we mock either cwuri or absolute_url, RDFGraphGenerator will add `sameAS` triples
        # on exported entities. We don't want this
        mock_same_as_uris.return_value = []
        with self.admin_access.client_cnx() as cnx:
            cat = cnx.entity_from_eid(self.cat_eid)
            with NamedTemporaryFile() as f:
                f.write(cat.view(action_regid))
                f.seek(0)
                graph = rdfio.default_graph()
                graph.load('file://' + f.name, rdf_format='xml')
            expected_graph = rdfio.default_graph()
            expected_graph.load('file://' + self.datapath(expected_fname), rdf_format='xml')
            iso1 = to_isomorphic(graph._graph)
            iso2 = to_isomorphic(expected_graph._graph)
            in_both, in_first, in_second = graph_diff(iso1, iso2)
            self.assertEqual(iso1, iso2, u'RDF graphs are not the same:\n'
                             '* Only in actual:\n{0}\n\n'
                             '* Only in expected:\n{1}'.format(
                                 sorted_triples(in_first), sorted_triples(in_second)))
        mock_download_url.assert_called_once_with()

    def test_complete_rdf_view(self):
        """Check that 'RDF export' action produce valid DCAT RDF data."""
        self._check_rdf_view('dcat.rdf.complete', 'valid_export.xml')

    def test_ckan_rdf_view(self):
        """Check bwcompat ckan views."""
        self._check_rdf_view('dcat.rdf.ckan', 'valid_export.xml')


class CkanJsonLicenseExportTC(testlib.CubicWebTC):
    """Test case for JSON license export to be used with CKAN."""

    configcls = PostgresApptestConfiguration

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            lic_scheme = conceptscheme(
                cnx, u'http://publications.europa.eu/resource/authority/licence')
            cnx.execute('SET CS scheme_relation RT WHERE CS eid %(cs)s, RT name %(rt)s',
                        {'cs': lic_scheme.eid, 'rt': 'license_type'})
            cnx.commit()

    def test_ckan_json_license_export_match_by_uri(self):
        """Check that a license concept scheme is exported correctly for CKAN (license found by its
        URI in OpenDefinition)."""
        with self.admin_access.repo_cnx() as cnx:
            lic_scheme = cnx.execute('ConceptScheme CS WHERE CS scheme_relation RT, '
                                     'RT name "license_type"').one()
            cc_by_40_uri = cnx.create_entity('ExternalUri',
                                             uri=u'http://creativecommons.org/licenses/by/4.0/')
            cc_by_40 = cnx.create_entity('Concept', in_scheme=lic_scheme,
                                         exact_match=cc_by_40_uri)
            cnx.create_entity('Label', label_of=cc_by_40,
                              label=u'Creative Commons Attribution 4.0 International',
                              kind=u"preferred", language_code=u'en')
            cnx.commit()
        with self.admin_access.web_request() as req:
            lic_scheme = req.entity_from_eid(lic_scheme.eid)
            rset = req.execute('Any X WHERE X eid %(x)s',
                               {'x': lic_scheme.eid})
            export = json.loads(req.view('dcat.ckan.json.licenses', rset=rset))
            self.assertEqual(len(export), 1)
            self.assertEqual(export[0], {  # Values from file data/od-licenses.json
                "domain_content": True,
                "domain_data": True,
                "domain_software": False,
                "family": "",
                "id": "CC-BY-4.0",
                "maintainer": "Creative Commons",
                "od_conformance": "approved",
                "osd_conformance": "not reviewed",
                "status": "active",
                "title": "Creative Commons Attribution 4.0 International",
                "url": "https://creativecommons.org/licenses/by/4.0/",
            })

    def test_ckan_json_license_export_match_by_title(self):
        """Check that a license concept scheme is exported correctly for CKAN (license found by its
        title in OpenDefinition)."""
        with self.admin_access.repo_cnx() as cnx:
            lic_scheme = cnx.execute('ConceptScheme CS WHERE CS scheme_relation RT, '
                                     'RT name "license_type"').one()
            pddl = cnx.create_entity('Concept', in_scheme=lic_scheme,
                                     cwuri=u'http://example.org/licenses/pddl/1.0/')
            cnx.create_entity('Label', label_of=pddl,
                              label=u'Open Data Commons Public Domain Dedication and License v1.0',
                              kind=u"preferred", language_code=u'en')
            cnx.commit()
        with self.admin_access.web_request() as req:
            rset = req.execute('Any X WHERE X eid %(x)s',
                               {'x': lic_scheme.eid})
            export = json.loads(req.view('dcat.ckan.json.licenses', rset=rset))
            self.assertEqual(len(export), 1)
            self.assertEqual(export[0], {  # Values from file data/od-licenses.json
                "domain_content": False,
                "domain_data": True,
                "domain_software": False,
                "family": "",
                "id": "ODC-PDDL-1.0",
                "maintainer": "",
                "od_conformance": "approved",
                "osd_conformance": "not reviewed",
                "status": "active",
                "title": "Open Data Commons Public Domain Dedication and License v1.0",
                "url": "http://www.opendefinition.org/licenses/odc-pddl",
            })

    def test_ckan_json_license_export_no_match(self):
        """Check that a license concept scheme is exported correctly for CKAN (license not found in
        OpenDefinition)."""
        with self.admin_access.repo_cnx() as cnx:
            lic_scheme = cnx.execute('ConceptScheme CS WHERE CS scheme_relation RT, '
                                     'RT name "license_type"').one()
            other_at = cnx.create_entity('Concept', in_scheme=lic_scheme)
            cnx.create_entity('Label', label_of=other_at, label=u'Licence Ouverte / Open Licence',
                              kind=u"preferred", language_code=u'en')
            cnx.commit()
        with self.admin_access.web_request() as req:
            licence = req.entity_from_eid(other_at.eid)
            rset = req.execute('Any X WHERE X eid %(x)s',
                               {'x': lic_scheme.eid})
            export = json.loads(req.view('dcat.ckan.json.licenses', rset=rset))
            self.assertEqual(len(export), 1)
            self.assertEqual(export[0], {
                "domain_content": False,
                "domain_data": True,
                "domain_software": False,
                "family": "",
                "maintainer": "",
                "od_conformance": "not reviewed",
                "osd_conformance": "not reviewed",
                "status": "active",
                "title": "Licence Ouverte / Open Licence",
                "id": licence.absolute_url(),
                "url": rdfio.permanent_url(licence),
            })

    def test_ckan_json_license_export_multiple_schemes(self):
        """Check several concept schemes can be exported with the view."""
        with self.admin_access.repo_cnx() as cnx:
            scheme1 = cnx.create_entity(
                'ConceptScheme', title=u'test scheme')
            cnx.execute('SET X scheme_relation RT'
                        ' WHERE RT name "license_type",'
                        ' X title "test"')
            scheme1.add_concept(u'foo')
            scheme2 = cnx.create_entity(
                'ConceptScheme', title=u'test 2')
            scheme2.add_concept(u'bar')
            cnx.execute('SET X scheme_relation RT'
                        ' WHERE RT name "license_type",'
                        ' X title LIKE "test%"')
            cnx.commit()
        with self.admin_access.web_request() as req:
            rset = req.execute(
                'Any CS WHERE CS is ConceptScheme, CS title LIKE "test%"')
            export = json.loads(req.view('dcat.ckan.json.licenses', rset=rset))
            self.assertCountEqual([x['title'] for x in export],
                                  [u'foo', u'bar'])


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-datacat tests for geoutils module."""

from cubicweb.devtools.testlib import BaseTestCase

from cubes.datacat.geoutils import (
    is_geometry, geometry,
    merge_in_geojson_collection,
    ogr_geom,
    srs_from_epsg,
)


class WorkingWithGeometryTC(BaseTestCase):
    """Test case for working with Geometry data."""

    def test_srs_from_epsg(self):
        """Check that we get a correct ``osr.SpatialReference`` from an EPSG code."""
        srs = srs_from_epsg(4326)
        self.assertEqual(srs.GetAttrValue('AUTHORITY', 0), 'EPSG')
        self.assertEqual(srs.GetAttrValue('AUTHORITY', 1), '4326')
        srs = srs_from_epsg('2154')
        self.assertEqual(srs.GetAttrValue('AUTHORITY', 0), 'EPSG')
        self.assertEqual(srs.GetAttrValue('AUTHORITY', 1), '2154')
        srs = srs_from_epsg(-1)
        self.assertIsNone(srs.GetAttrValue('AUTHORITY', 0))
        self.assertIsNone(srs.GetAttrValue('AUTHORITY', 1))
        with self.assertRaises(ValueError) as cm:
            srs_from_epsg('abcd')
        self.assertIn('Invalid EPSG', str(cm.exception))

    def test_ogr_geom(self):
        """Check that we get correct a ``ogr.Geometry`` from a given value."""
        # (x, y) value
        geom = ogr_geom((2, 46))
        self.assertEqual(geom.ExportToWkt(), 'POINT (2 46)')
        # WKT value
        geom = ogr_geom('POINT (2 46)')
        self.assertEqual(geom.ExportToWkt(), 'POINT (2 46)')
        # GeoJSON value
        geom = ogr_geom('{"type": "Point", "coordinates": [2.0, 46.0]}')
        self.assertEqual(geom.ExportToWkt(), 'POINT (2 46)')
        # Invalid value
        with self.assertRaises(ValueError) as cm:
            ogr_geom('POINT (2, 46)')
        self.assertIn('Invalid value for Geometry', str(cm.exception))

    def test_is_geometry(self):
        """Check that we can determine if geometry value is valid for cubicweb-postgis."""
        # When valid WKT
        self.assertTrue(is_geometry('POINT (2 46)'))
        # When invalid WKT
        self.assertFalse(is_geometry('POINT (2, 46)'))
        # When not WKT
        self.assertFalse(is_geometry('{"type": "Point", "coordinates": [2.0, 46.0]}'))
        # When geom types are the same
        self.assertTrue(is_geometry('POINT (2 46)', expected_geom_type='Point'))
        # When geom types are not the same
        self.assertFalse(is_geometry('POINT (2 46)', expected_geom_type='MultiPoint'))
        # When epsgs are the same
        self.assertTrue(is_geometry('POINT (2 46)', epsg=4326, expected_epsg=4326))
        # When epsgs are not the same
        self.assertFalse(is_geometry('POINT (2 46)', epsg=4326, expected_epsg=2154))
        # When epsg is expected but not given
        self.assertFalse(is_geometry('POINT (2 46)', expected_epsg=4326))
        # When epsg is not integer
        self.assertFalse(is_geometry('POINT (2 46)', epsg='abcd', expected_epsg='abcd'))
        # When invalid geometry (eg. not closed polygon)
        self.assertFalse(is_geometry('POLYGON ((30 10, 40 40, 20 40, 10 20))'))

    def test_convert_geometry(self):
        """Check that we get correct Geometry value from given value."""
        # When value is lat/long
        self.assertEqual(geometry((2, 46)), ('POINT (2 46)', 0))
        # When EPSG is given
        self.assertEqual(geometry((2, 46), epsg=4326), ('POINT (2 46)', 4326))
        # When expected geom type is the same
        value = 'MULTIPOINT (2 46,2.5 45.5)'
        self.assertEqual(geometry(value, expected_geom_type='MultiPoint'), (value, 0))
        # Point -> MultiPoint
        self.assertEqual(geometry('POINT (2.0 46.0)', epsg=4326, expected_geom_type='MultiPoint'),
                         ('MULTIPOINT (2 46)', 4326))
        # LineString -> MultiLineString
        self.assertEqual(geometry('LINESTRING (30 10, 10 30, 40 40)', epsg=4326,
                                  expected_geom_type='MultiLineString'),
                         ('MULTILINESTRING ((30 10,10 30,40 40))', 4326))
        # Polygon -> MultiPolygon
        self.assertEqual(geometry('POLYGON ((30 10, 40 40, 20 40, 10 20, 30 10))', epsg=4326,
                                  expected_geom_type='MultiPolygon'),
                         ('MULTIPOLYGON (((30 10,40 40,20 40,10 20,30 10)))', 4326))
        # -> GeometryCollection
        value = 'POINT (2 46)'
        self.assertEqual(geometry(value, epsg=4326, expected_geom_type='GeometryCollection'),
                         ('GEOMETRYCOLLECTION ({})'.format(value), 4326))
        # Reprojection
        self.assertEqual(geometry('POINT (2 46)', epsg=4326, expected_epsg=2154),
                         ('POINT (622609.119992160820402 6544963.910583730787039)', 2154))
        # Reprojection and type tranformation
        self.assertEqual(geometry('POINT (2 46)', epsg=4326,
                                  expected_geom_type='multipoint', expected_epsg=2154),
                         ('MULTIPOINT (622609.119992160820402 6544963.910583730787039)', 2154))

    def test_cannot_convert_geometry(self):
        """Check that in some cases a ValueError is raised."""
        # When value is invalid WKT
        with self.assertRaises(ValueError) as cm:
            geometry('POINT (2.0, 46.0)')
        self.assertIn('Invalid value for Geometry', str(cm.exception))
        # When geom_types are incompatible
        with self.assertRaises(ValueError) as cm:
            geometry('LINESTRING (2.0 46.0,2.5 45.5)', expected_geom_type='MultiPolygon')
        self.assertIn('Cannot convert LINESTRING', str(cm.exception))
        self.assertIn('to MULTIPOLYGON', str(cm.exception))
        # When invalid epsg
        with self.assertRaises(ValueError) as cm:
            geometry('POINT (2 46)', epsg='abcd')
        self.assertIn('Invalid EPSG', str(cm.exception))
        with self.assertRaises(ValueError) as cm:
            geometry('POINT (2 46)', epsg=2154, expected_epsg='abcd')
        self.assertIn('Invalid EPSG', str(cm.exception))
        # When missing origin epsg
        with self.assertRaises(ValueError) as cm:
            geometry('POINT (2 46)', expected_epsg=2154)
        self.assertIn('Invalid or missing origin EPSG', str(cm.exception))

    def test_convert_geometry_from_geojson(self):
        """Check that we get correct geometry value from GeoJSON."""
        value = '{"type": "Point", "coordinates": [2.0, 46.0]}'
        self.assertEqual(geometry(value, expected_geom_type='MultiPoint'),
                         ('MULTIPOINT (2 46)', 0))
        # Embbed srs
        value = ('{"type": "Point", "coordinates": [2.0, 46.0], '
                 '"crs": {"type": "name", "properties": {"name": "EPSG:4326"}}}')
        self.assertEqual(geometry(value), ('POINT (2 46)', 4326))
        self.assertEqual(geometry(value, expected_geom_type='MultiPoint', expected_epsg=2154),
                         ('MULTIPOINT (622609.119992160820402 6544963.910583730787039)', 2154))


class WorkingWithGeoJSONTC(BaseTestCase):
    """Test case for working with GeoJSON objects."""

    def test_merge_in_geojson_collection(self):
        """Check that merging GeoJSON features return correct feature collection."""
        # One empty object
        geojson = {}
        actual = merge_in_geojson_collection([geojson])
        self.assertEqual(actual, {})
        self.assertIsNot(actual, geojson)
        # One Feature object
        geojson = {'type': u'Feature', 'properties': {'name': u'France'},
                   'geometry': {'type': u'Point', 'coordinates': [2, 46]}}
        expected = {'type': u'FeatureCollection', 'features': [geojson]}
        actual = merge_in_geojson_collection([geojson])
        self.assertEqual(actual, expected)
        # One FeatureCollection object
        geojson = {'type': u'FeatureCollection', 'features': [
            {'type': u'Feature', 'properties': {'name': u'France'},
             'geometry': {'type': u'Point', 'coordinates': [2, 46]}}
        ]}
        actual = merge_in_geojson_collection([geojson])
        self.assertEqual(actual, geojson)
        self.assertIsNot(actual, geojson)
        # Two objects, both empty
        geojson1 = {}
        geojson2 = {}
        actual = merge_in_geojson_collection([geojson1, geojson2])
        self.assertEqual(actual, {})
        self.assertIsNot(actual, geojson1)
        self.assertIsNot(actual, geojson2)
        # Two objects, one empty
        geojson1 = {}
        geojson2 = {'type': u'Feature', 'properties': {'name': u'France'},
                    'geometry': {'type': u'Point', 'coordinates': [2, 46]}}
        expected = {'type': u'FeatureCollection', 'features': [geojson2]}
        actual = merge_in_geojson_collection([geojson1, geojson2])
        self.assertEqual(actual, expected)
        # Two objects, other empty
        geojson1 = {'type': u'Feature', 'properties': {'name': u'France'},
                    'geometry': {'type': u'Point', 'coordinates': [2, 46]}}
        geojson2 = {}
        expected = {'type': u'FeatureCollection', 'features': [geojson1]}
        actual = merge_in_geojson_collection([geojson1, geojson2])
        self.assertEqual(actual, expected)
        # Feature with feature
        geojson1 = {'type': u'Feature', 'properties': {'name': u'France'},
                    'geometry': {'type': u'Point', 'coordinates': [2, 46]}}
        geojson2 = {'type': u'Feature', 'properties': {'name': u'France'},
                    'geometry': {'type': u'Point', 'coordinates': [3, 47]}}
        expected = {'type': u'FeatureCollection', 'features': [geojson1, geojson2]}
        actual = merge_in_geojson_collection([geojson1, geojson2])
        self.assertEqual(actual, expected)
        # FeatureCollection with Feature
        geojson1 = {'type': u'FeatureCollection', 'features': [
            {'type': u'Feature', 'properties': {'name': u'France'},
             'geometry': {'type': u'Point', 'coordinates': [2, 46]}}
        ]}
        geojson2 = {'type': u'Feature', 'properties': {'name': u'France'},
                    'geometry': {'type': u'Point', 'coordinates': [3, 47]}}
        features = geojson1['features']
        expected = {'type': 'FeatureCollection', 'features': features + [geojson2]}
        actual = merge_in_geojson_collection([geojson1, geojson2])
        self.assertEqual(actual['type'], expected['type'])
        self.assertCountEqual(actual['features'], expected['features'])
        # Feature with FeatureCollection
        geojson1 = {'type': u'Feature', 'properties': {'name': u'France'},
                    'geometry': {'type': u'Point', 'coordinates': [2, 46]}}
        geojson2 = {'type': u'FeatureCollection', 'features': [
            {'type': u'Feature', 'properties': {'name': u'France'},
             'geometry': {'type': u'Point', 'coordinates': [3, 47]}}
        ]}
        features = geojson2['features']
        expected = {'type': 'FeatureCollection', 'features': features + [geojson1]}
        actual = merge_in_geojson_collection([geojson1, geojson2])
        self.assertEqual(actual['type'], expected['type'])
        self.assertCountEqual(actual['features'], expected['features'])
        # FeatureCollection with FeatureCollection
        geojson1 = {'type': u'FeatureCollection', 'features': [
            {'type': u'Feature', 'properties': {'name': u'France'},
             'geometry': {'type': u'Point', 'coordinates': [2, 46]}}
        ]}
        geojson2 = {'type': u'FeatureCollection', 'features': [
            {'type': u'Feature', 'properties': {'name': u'France'},
             'geometry': {'type': u'Point', 'coordinates': [3, 47]}}
        ]}
        features1 = geojson1['features']
        features2 = geojson2['features']
        expected = {'type': 'FeatureCollection', 'features': features1 + features2}
        actual = merge_in_geojson_collection([geojson1, geojson2])
        self.assertEqual(actual['type'], expected['type'])
        self.assertCountEqual(actual['features'], expected['features'])


if __name__ == '__main__':
    from unittest import main
    main()



"""cubicweb-datacat test utilities"""

from cubicweb import Binary


def create_file(cnx, data, data_name=None, **kwargs):
    """Create a File entity"""
    data_name = data_name or data.decode('utf-8')
    kwargs.setdefault('data_format', u'text/plain')
    return cnx.create_entity('File', data=Binary(data),
                             data_name=data_name,
                             **kwargs)


def produce_file(cnx, resourcefeed, inputfile):
    """Simulate the production of `inputfile` by resource feed `resourcefeed`"""
    tseq = resourcefeed.transformation_script[0].transformation_sequence
    # Build a transformation process "by hand".
    with cnx.security_enabled(write=False):
        process = cnx.create_entity(
            'DataTransformationProcess',
            process_input_file=inputfile,
            process_for_resourcefeed=resourcefeed,
            transformation_sequence=tseq)
        cnx.commit()  # transformation triggered in a hook.
    return process.reverse_produced_by[0]


def mediatypes_scheme(cnx, *labels):
    """Build a "dummy" media-types concept scheme to satisfy hooks on
    distribution/file media type conformance. Return concepts corresponding to
    requested `labels`.
    """
    scheme = cnx.create_entity(
        'ConceptScheme',
        cwuri=u'http://www.iana.org/assignments/media-types/media-types.xml')
    return [scheme.add_concept(label) for label in labels]



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-datacat command plugins"""

from __future__ import print_function

import io
import json
import os
import os.path as osp
import requests
import zipfile

from logilab.common.shellutils import ProgressBar

from cubicweb.cwctl import CWCTL
from cubicweb.dataimport.importer import SimpleImportLog
from cubicweb.toolsutils import (
    Command,
    underline_title,
)
from cubicweb.utils import admincnx

from cubes.datacat import cwsource_pull_data
from cubes.datacat.vocabularies import (
    ADMS_FRENCH_LABELS,
    eurovoc_domains,
    EUROVOC_URI,
    LICENSE_SCHEME,
    simplify_eurovoc,
)
from cubes.skos import rdfio
from cubes.skos.ccplugin import ImportSkosData
from cubes.skos.sobjects import (
    graph_extentities,
    import_skos_extentities,
)


class SyncSKOSSchemes(Command):
    """Synchronize SKOS schemes for a datacat instance.

    <instance>
      identifier of a datacat instance.

    """
    arguments = '<instance>'
    name = 'sync-schemes'
    min_args = 1
    max_args = 1

    def run(self, args):
        appid = args[0]
        with admincnx(appid) as cnx:
            rset = cnx.execute(
                'Any S ORDERBY X WHERE X is SKOSSource, X through_cw_source S')
            title = '-> synchronizing SKOS sources'
            pb = ProgressBar(len(rset), title=title)
            created, updated = set([]), set([])
            for eid, in rset:
                stats = cwsource_pull_data(cnx.repo, eid, raise_on_error=False)
                pb.update()
                created.update(stats['created'])
                updated.update(stats['updated'])
            # French translations
            qs = (u'Any C WHERE C cwuri IN ({0}), '
                  u'NOT EXISTS(L label_of C, L kind "preferred", L language_code "fr")'.format(
                      ', '.join([u"'{0}'".format(uri) for uri in ADMS_FRENCH_LABELS])))
            uri2eid = dict((concept.cwuri, concept.eid) for concept in cnx.execute(qs).entities())
            title = '-> Adding missing french labels for ADMS concepts'
            pb = ProgressBar(len(uri2eid), title=title)
            for concept_uri, concept_eid in uri2eid.items():
                fr_label = ADMS_FRENCH_LABELS.get(concept_uri)
                if fr_label:
                    label = cnx.create_entity('Label', label=fr_label, language_code=u'fr',
                                              kind=u'preferred', label_of=concept_eid)
                    created.add(label)
                pb.update()
            cnx.commit()
            print('\n   {0} created, {1} updated'.format(len(created), len(updated)))
            print('\n-> Adding constraints for license schemes')
            cnx.execute('SET CS scheme_relation RT WHERE CS cwuri %(scheme_uri)s, '
                        'RT name %(rtype_name)s',
                        {'scheme_uri': LICENSE_SCHEME, 'rtype_name': 'license_type'})
            cnx.commit()


def set_store_default(cls):
    """Set default value for "cw-store" option to "massive"."""
    for option in cls.options:
        if option[0] == 'cw-store':
            option[1]['default'] = 'massive'
            break
    return cls


@set_store_default
class ImportEurovoc(ImportSkosData):
    """Synchronize Eurovoc concept schemes for a datacat instance.

    <instance>
      identifier of a datacat instance.

    """
    arguments = '[options] <instance>'
    name = 'import-eurovoc'
    min_args = 1
    options = (ImportSkosData.options + (
        ('update', {
            'short': 'u', 'type': 'yn', 'default': False,
            'help': 'Force Eurovoc thesaurus download, even file already exists'
        }),
    ))

    def run(self, args):
        appid = args[0]
        update = self.get('update')
        with admincnx(appid) as cnx:
            datadir = osp.join(cnx.repo.config.appdatahome, 'eurovoc')
        if not osp.isdir(datadir):
            os.mkdir(datadir)
        zip_fname = osp.join(datadir, 'eurovoc_skos.zip')
        n3_fname = osp.join(datadir, 'eurovoc_skos.n3')
        simplified_fname = osp.join(datadir, 'eurovoc_skos_simple.n3')
        json_fname = osp.join(datadir, 'eurovoc_domains.json')
        print(u'\n{0}'.format(underline_title('Importing Eurovoc thesaurus')))
        if not osp.isfile(zip_fname) or update:
            print(u'Downloading eurovoc SKOS/RDF ZIP file from publications.europa.eu')
            r = requests.get(EUROVOC_URI, stream=True)
            with open(zip_fname, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192):
                    if chunk:
                        f.write(chunk)
        elif osp.isfile(zip_fname):
            print(u'Using existing eurovoc SKOS/RDF ZIP file {0}'.format(zip_fname))
        if not osp.isfile(n3_fname) or update:  # n3 file is faster to load
            print(u'Extracting RDF file and converting to Turtle')
            eurovoc_graph = rdfio.default_graph()
            with zipfile.ZipFile(zip_fname) as eurovoc_zip:
                eurovoc_graph.load(eurovoc_zip.open('eurovoc_skos.rdf'), rdf_format='xml')
                with open(n3_fname, 'w') as f:
                    eurovoc_graph.dump(f, rdf_format='n3')
        print(u'Preparing data for import')
        simplified_graph = rdfio.default_graph()
        if not osp.isfile(simplified_fname) or update:
            eurovoc_graph = rdfio.default_graph()
            with open(n3_fname) as f:
                eurovoc_graph.load(f, rdf_format='n3')
            # Flatten and simplify the graph
            simplify_eurovoc(eurovoc_graph, simplified_graph)
            with open(simplified_fname, 'w') as f:
                simplified_graph.dump(f, rdf_format='n3')
        else:
            with open(simplified_fname) as f:
                simplified_graph.load(f, rdf_format='n3')
        if not osp.isfile(json_fname) or update:
            with io.open(json_fname, 'w', encoding='utf-8') as f:
                f.write(unicode(json.dumps(eurovoc_domains(simplified_graph), ensure_ascii=False,
                                           sort_keys=True, indent=2, separators=(u',', u':'))))
            print(u'-> JSON file with Eurovoc domains created at {0}'.format(json_fname))
        print(u'Importing data')
        with admincnx(appid) as cnx:
            import_log = SimpleImportLog('eurovoc.import')
            cwsource = cnx.execute(u'Any S WHERE S is CWSource, S url %(eurovoc_uri)s',
                                   {'eurovoc_uri': EUROVOC_URI}).one()
            store = self.cw_store_factories[self.get('cw-store')](cnx)
            (created, updated), conceptschemes = import_skos_extentities(
                cnx, graph_extentities(simplified_graph), import_log, source=cwsource, store=store)
            cnx.commit()
            print(u'Created: {0}\nUpdated: {0}'.format(len(created), len(updated)))
            if conceptschemes:
                print(u'Concept schemes:\n* {0}'.format('\n* '.join(conceptschemes)))
            _update_source(cnx, cwsource)
            cnx.repo.shutdown()


def _update_source(cnx, cwsource):
    """Update cw_source relation of entities to point to Eurovoc CWSource.
    """
    sql = cnx.system_sql
    sqlqs = ("UPDATE cw_source_relation SET eid_to=%(source)s WHERE eid_from in ("
             "SELECT _X.cw_eid FROM cw_{0} AS _X WHERE _X.cw_cwuri ILIKE '%%eurovoc%%')")
    for tablename in ('Label', 'Concept', 'ConceptScheme'):
        sql(sqlqs.format(tablename), {'source': cwsource.eid})
    cnx.commit()


class CleanResourceFeedProcesses(Command):
    """Remove ancient resource feed's processes.

    <instance>
      identifier of a datacat instance.
    """
    arguments = '<instance>'
    name = 'clean-processes'
    min_args = 1
    max_args = 1

    def run(self, args):
        appid = args[0]
        with admincnx(appid) as cnx:
            deleted = cnx.call_service('datacat.clean-resourcefeed-process-logs')
            cnx.commit()
        if deleted:
            print('deleted %d process logs' % len(deleted))
        else:
            print('nothing deleted')


for cmd in (SyncSKOSSchemes, ImportEurovoc, CleanResourceFeedProcesses):
    CWCTL.register(cmd)



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import urllib2
import StringIO
import os.path

from logilab.common.decorators import monkeypatch
from logilab.database import FunctionDescr
from rql.utils import register_function
from cubicweb.server.sources import datafeed
from cubicweb.xy import xy

from cubes import skos


xy.register_prefix('dct', 'http://purl.org/dc/terms/')
xy.register_prefix('adms', 'http://www.w3.org/ns/adms#')
xy.register_prefix('dcat', 'http://www.w3.org/ns/dcat#')
xy.register_prefix('vcard', 'http://www.w3.org/2006/dcat/ns#')
xy.add_equivalence('Agent name', 'foaf:name')
xy.add_equivalence('Agent name', 'vcard:fn')  # XXX overrides ^?
xy.add_equivalence('Agent email', 'vcard:hasEmail')
xy.add_equivalence('Dataset identifier', 'dct:identifier')
xy.add_equivalence('Dataset title', 'dct:title')
xy.add_equivalence('Dataset creation_date', 'dct:issued')
xy.add_equivalence('Dataset modification_date', 'dct:modified')
xy.add_equivalence('Dataset description', 'dct:description')
xy.add_equivalence('Dataset keyword', 'dcat:keyword')
xy.add_equivalence('Dataset keyword', 'dcat:theme')
xy.add_equivalence('Dataset landing_page', 'dcat:landingPage')
xy.add_equivalence('Dataset spatial_coverage', 'dct:spatial')
xy.add_equivalence('Dataset provenance', 'dct:provenance')
xy.add_equivalence('Dataset dataset_distribution', 'dcat:distribution')
xy.add_equivalence('Dataset dataset_publisher', 'dct:publisher')
xy.add_equivalence('Dataset dataset_contact_point', 'adms:contactPoint')
xy.add_equivalence('Distribution access_url', 'dcat:accessURL')
xy.add_equivalence('Distribution description', 'dct:description')
xy.add_equivalence('Distribution format', 'dct:format')
xy.add_equivalence('Distribution licence', 'dct:license')


# DataFeedParser.retrieve_url fix to support ftp:// URLs and local files.
@monkeypatch(datafeed.DataFeedParser)
def retrieve_url(self, url, data=None, headers=None):
    """Return stream linked by the given url:
    * HTTP urls will be normalized (see :meth:`normalize_url`)
    * handle file:// URL
    * other will be considered as plain content, useful for testing purpose
    """
    headers = headers or {}
    if url.startswith(('http', 'ftp')):
        url = self.normalize_url(url)
        if url.startswith('http') and data:
            self.source.info('POST %s %s', url, data)
        else:
            self.source.info('GET %s', url)
        req = urllib2.Request(url, data, headers)
        return datafeed._OPENER.open(req, timeout=self.source.http_timeout)
    if url.startswith('file://'):
        return datafeed.URLLibResponseAdapter(open(url[7:]), url)
    if os.path.isfile(url):
        return datafeed.URLLibResponseAdapter(open(url), url)
    return datafeed.URLLibResponseAdapter(StringIO.StringIO(url), url)


# Needed until https://www.cubicweb.org/ticket/11257307 is done
@monkeypatch(skos)
def register_skos_rdf_list_mapping(reg):
    """Register minimal SKOS mapping, only describing concept scheme (for use in 'list.rdf' views).
    """
    reg.register_prefix('dct', 'http://purl.org/dc/terms/')
    reg.register_prefix('skos', 'http://www.w3.org/2004/02/skos/core#')
    reg.register_etype_equivalence('ConceptScheme', 'skos:ConceptScheme')
    reg.register_attribute_equivalence('ConceptScheme', 'title', 'dct:title')
    reg.register_attribute_equivalence('ConceptScheme', 'description', 'dct:description')


# Register new Postgis functions

class ST_COLLECTIONEXTRACT(FunctionDescr):
    minargs = 2
    maxargs = 2
    supported_backends = ('postgres',)
    rtype = 'Geometry'


class ST_GEOMFROMGEOJSON(FunctionDescr):
    aggregat = True
    minargs = 1
    maxargs = 1
    supported_backends = ('postgres',)
    rtype = 'Geometry'


register_function(ST_COLLECTIONEXTRACT)
register_function(ST_GEOMFROMGEOJSON)



"""cubicweb-datacat application package

Data catalog
"""

import json
from os.path import (
    dirname,
    join,
)


def register_catalog_rdf_mapping(reg):
    """Register mapping between DCAT RDF vocabulary and DataCatalog entity type."""
    reg.register_prefix('dcat', 'http://www.w3.org/ns/dcat#')
    reg.register_prefix('dcterms', 'http://purl.org/dc/terms/')
    reg.register_etype_equivalence('DataCatalog', 'dcat:Catalog')
    reg.register_attribute_equivalence('DataCatalog', 'title', 'dcterms:title')
    reg.register_attribute_equivalence('DataCatalog', 'description', 'dcterms:description')
    reg.register_attribute_equivalence('DataCatalog', 'issued', 'dcterms:issued')
    reg.register_attribute_equivalence('DataCatalog', 'modified', 'dcterms:modified')
    reg.register_relation_equivalence('DataCatalog', 'catalog_publisher', 'Agent',
                                      'dcterms:publisher')
    reg.register_relation_equivalence('Dataset', 'in_catalog', 'DataCatalog', 'dcat:dataset',
                                      reverse=True)


def register_dataset_rdf_mapping(reg):
    """Register mapping between DCAT RDF vocabulary and Dataset entity type."""
    reg.register_prefix('dcat', 'http://www.w3.org/ns/dcat#')
    reg.register_prefix('dcterms', 'http://purl.org/dc/terms/')
    reg.register_etype_equivalence('Dataset', 'dcat:Dataset')
    reg.register_attribute_equivalence('Dataset', 'title', 'dcterms:title')
    reg.register_attribute_equivalence('Dataset', 'description', 'dcterms:description')
    reg.register_attribute_equivalence('Dataset', 'issued', 'dcterms:issued')
    reg.register_attribute_equivalence('Dataset', 'modified', 'dcterms:modified')
    reg.register_attribute_equivalence('Dataset', 'identifier', 'dcterms:identifier')
    reg.register_attribute_equivalence('Dataset', 'landing_page', 'dcat:landingPage')
    reg.register_relation_equivalence('Dataset', 'in_catalog', 'DataCatalog', 'dcat:dataset',
                                      reverse=True)
    reg.register_relation_equivalence('Dataset', 'dataset_publisher', 'Agent',
                                      'dcterms:publisher')
    reg.register_relation_equivalence('Dataset', 'dataset_contact_point', 'Agent',
                                      'dcat:contactPoint')
    reg.register_relation_equivalence('Dataset', 'temporal_coverage', 'PeriodOfTime',
                                      'dcterms:temporal')
    reg.register_relation_equivalence('Distribution', 'of_dataset', 'Dataset', 'dcat:distribution',
                                      reverse=True)


def register_periodoftime_rdf_mapping(reg):
    """Register mapping between DCAT RDF vocabulary and PeriodOfTime entity
    type.
    """
    reg.register_prefix('dcterms', 'http://purl.org/dc/terms/')
    reg.register_prefix('schema', 'http://schema.org/')
    reg.register_etype_equivalence('PeriodOfTime', 'dcterms:PeriodOfTime')
    reg.register_attribute_equivalence('PeriodOfTime', 'start', 'schema:startDate')
    reg.register_attribute_equivalence('PeriodOfTime', 'end', 'schema:endDate')


def register_distribution_rdf_mapping(reg):
    """Register mapping between DCAT RDF vocabulary and Distribution entity type."""
    reg.register_prefix('dcat', 'http://www.w3.org/ns/dcat#')
    reg.register_prefix('dcterms', 'http://purl.org/dc/terms/')
    reg.register_etype_equivalence('Distribution', 'dcat:Distribution')
    reg.register_attribute_equivalence('Distribution', 'title', 'dcterms:title')
    reg.register_attribute_equivalence('Distribution', 'description', 'dcterms:description')
    reg.register_attribute_equivalence('Distribution', 'issued', 'dcterms:issued')
    reg.register_attribute_equivalence('Distribution', 'modified', 'dcterms:modified')
    reg.register_attribute_equivalence('Distribution', 'byte_size', 'dcat:byteSize')
    reg.register_relation_equivalence('Distribution', 'of_dataset', 'Dataset', 'dcat:distribution',
                                      reverse=True)


def register_publisher_rdf_mapping(reg):
    """Register mapping between FOAF RDF vocabulary and Agent entity type."""
    reg.register_prefix('foaf', 'http://xmlns.com/foaf/0.1/')
    reg.register_prefix('dcterms', 'http://purl.org/dc/terms/')
    reg.register_etype_equivalence('Agent', 'foaf:Agent')
    reg.register_attribute_equivalence('Agent', 'name', 'foaf:name')
    reg.register_relation_equivalence('Dataset', 'dataset_publisher', 'Agent',
                                      'dcterms:publisher')


def register_contact_point_rdf_mapping(reg):
    """Register mapping between VCard RDF vocabulary and Agent entity type."""
    reg.register_prefix('vcard', 'http://www.w3.org/2006/vcard/ns#')
    reg.register_prefix('dcat', 'http://www.w3.org/ns/dcat#')
    reg.register_relation_equivalence('Dataset', 'dataset_contact_point', 'Agent',
                                      'dcat:contactPoint')
    reg.register_etype_equivalence('Agent', 'vcard:Kind')
    reg.register_attribute_equivalence('Agent', 'name', 'vcard:fn')


def register_location_rdf_mapping(reg):
    """Register mapping between DCAT RDF vocabulary and Location entity type."""
    reg.register_prefix('dcterms', 'http://purl.org/dc/terms/')
    reg.register_prefix('locn', 'http://www.w3.org/ns/locn#')
    reg.register_etype_equivalence('Location', 'dcterms:Location')
    reg.register_relation_equivalence('Location', 'locate_datasets', 'Dataset',
                                      'dcterms:spatial', reverse=True)


def concept_in(cnx, label, scheme_uri):
    """Return a concept with the given label and in the given concept scheme."""
    rset = cnx.execute('Any C WHERE C in_scheme S, S cwuri %(scheme_uri)s, '
                       'L label_of C, L kind "preferred", L label %(label)s',
                       {'label': label, 'scheme_uri': scheme_uri})
    if rset:
        return rset.one()
    else:
        return None


def cwsource_pull_data(repo, cwsource_eid, **kwargs):
    """Pull data from a CWSource associated with a ResourceFeed"""
    dfsource = repo.sources_by_eid[cwsource_eid]
    kwargs.setdefault('force', True)
    kwargs.setdefault('raise_on_error', True)
    with repo.internal_cnx() as icnx:
        stats = dfsource.pull_data(icnx, **kwargs)
        icnx.commit()
    return stats


def opendefinition_licenses():
    """Return a dict with all licences' definition as provided by
    http://licenses.opendefinition.org/.
    """
    od_path = join(dirname(__file__), 'data', 'od_licenses.json')
    with open(od_path) as f:
        return json.load(f)



# copyright 2014-2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-datacat server objects"""

import hashlib

from cubicweb.predicates import match_user_groups
from cubicweb.server import Service
from cubicweb.server.sources import datafeed
from cubicweb.dataimport.stores import (
    MetadataGenerator,
    NoHookRQLObjectStore,
)
from cubicweb.dataimport.importer import (
    ExtEntitiesImporter,
    ExtEntity,
)


def compute_sha1hex(value):
    """Return the SHA1 hex digest of `value`."""
    return unicode(hashlib.sha1(value).hexdigest())


class ResourceFeedParser(datafeed.DataFeedParser):
    """Fetch files and apply validation and transformation processes before
    attaching them to their Dataset as Resources.
    """
    __regid__ = 'datacat.resourcefeed-parser'

    def resourcefeed(self):
        cwsource = self._cw.entity_from_eid(self.source.eid)
        if cwsource.reverse_resource_feed_source:
            return cwsource.reverse_resource_feed_source[0]

    def importer(self, url, extid2eid):
        """Return an ExtEntity importer."""
        cnx = self._cw
        schema = cnx.vreg.schema
        # XXX Using NoHookRQLObjectStore because RQLObjectStore does not
        # support setting cw_source relation.
        metagenerator = MetadataGenerator(cnx, source=self.source)
        store = NoHookRQLObjectStore(cnx, metagenerator)
        return ExtEntitiesImporter(schema, store, import_log=self.import_log,
                                   extid2eid=extid2eid)

    def extid2eid(self):
        qs = 'Any H,X WHERE X data_sha1hex H, X cw_source S, S eid %(s)s'
        args = {'s': self.source.eid}
        return dict(self._cw.execute(qs, args))

    def process(self, url, raise_on_error=False):
        """Build a File entity from data fetched from url"""
        resourcefeed = self.resourcefeed()
        if resourcefeed is None:
            self.error('parser "%s" cannot be used for a CWSource with a ResourceFeed attached',
                       self.__regid__)
            if raise_on_error:
                raise Exception('CWSource #%d has no ResourceFeed attached' % self.source.eid)
        stream = self.retrieve_url(url)
        data = stream.read()
        sha1hex = compute_sha1hex(data)
        self.import_log.record_info(
            'fetched data from {0} sha1={1}'.format(url, sha1hex))
        extid2eid = self.extid2eid()
        extentity = ExtEntity('File', sha1hex)
        if extentity.extid not in extid2eid:
            # Only set `values` when the entity does not already exist
            # (otherwise the importer will consider the entity for update
            # which is not what we want).
            extentity.values.update({
                'data_name': set([url.split('/')[-1]]),
                'data_sha1hex': set([sha1hex]),
                'data': set([data]),
                'cwuri': set([sha1hex]),
            })
        importer = self.importer(url, extid2eid)
        importer.import_entities([extentity])
        created, updated = importer.created, importer.updated
        msg, args = ('data import for %s completed: created %d entities, updated %s entities',
                     (url, len(created), len(updated)))
        self.info(msg, *args)
        self.import_log.record_info(msg % args)
        entity = None
        if created:
            assert not updated, 'unexpected update of entities {0}'.format(updated)
            assert len(created) == 1, created
            entity = self._cw.entity_from_eid(created.pop())
            self.stats['created'].add(entity.eid)
        if updated:
            assert not created, 'unexpected creation of entities {0}'.format(created)
            assert len(updated) == 1, updated
            entity = self._cw.entity_from_eid(updated.pop())
            self.stats['updated'].add(entity.eid)
        if entity is not None:
            # Ensure imported entity as a cw_source, otherwise re-import would
            # not work.
            assert entity.cw_source
            self.process_data(entity)

    def process_data(self, entity):
        """Launch validation and transformation scripts of all related
        ResourceFeed entities.
        """
        resourcefeed = self.resourcefeed()
        distribution = resourcefeed.resourcefeed_distribution[0]
        relations = {
            'data_format': resourcefeed.data_format,
            'file_distribution': distribution,
        }
        self.import_log.record_debug(
            'adding file #{0} to distribution #{1}'.format(
                entity.eid, distribution.eid))
        if distribution.reverse_file_distribution:
            distfile = distribution.reverse_file_distribution[0]
            self.import_log.record_debug(
                'file #{0} replaces previous distribution file #{1}'.format(
                    entity.eid, distfile.eid))
            relations['replaces'] = distfile
        entity.cw_set(**relations)

        # Launch validation/transformation processes.
        vprocess = None
        if resourcefeed.validation_script:
            vscript = resourcefeed.validation_script[0]
            vprocess = self._cw.create_entity('DataValidationProcess',
                                              process_for_resourcefeed=resourcefeed,
                                              validation_script=vscript,
                                              process_input_file=entity)
            msg, args = ('created validation process #%d for file #%d',
                         (vprocess.eid, entity.eid))
            self.info(msg, *args)
            self.import_log.record_info(msg % args)
        if resourcefeed.transformation_script:
            tseq = resourcefeed.transformation_script[0].transformation_sequence
            tprocess = self._cw.create_entity('DataTransformationProcess',
                                              process_for_resourcefeed=resourcefeed,
                                              transformation_sequence=tseq,
                                              process_input_file=entity,
                                              process_depends_on=vprocess)
            msg, args = ('created transformation process #%d for file #%d',
                         (tprocess.eid, entity.eid))
            self.info(msg, *args)
            self.import_log.record_info(msg % args)


def delete_process_logs_for(cnx, eid):
    """Delete old process log files for ResourceFeed with `eid`."""
    keep = cnx.entity_from_eid(eid).processes_to_keep
    process_etypes = cnx.vreg.schema['process_for_resourcefeed'].targets(
        'ResourceFeed', 'object')
    deleted = []
    for etype in process_etypes:
        rset = cnx.execute(
            'Any X ORDERBY X DESC LIMIT {limit}'
            ' WHERE X process_for_resourcefeed R, R eid %(eid)s,'
            ' X is {etype}'.format(etype=etype, limit=keep),
            {'eid': eid})
        if not rset:
            continue
        last_to_keep = rset[-1][0]
        qs = ('DELETE File F WHERE P process_stderr F, P is {etype}, P eid < {last},'
              ' P process_for_resourcefeed R, R eid %(eid)s').format(
                  etype=etype, last=last_to_keep)
        rset = cnx.execute(qs, {'eid': eid})
        deleted.extend([x for x, in rset])
    return deleted


class RemoveOldProcessesService(Service):
    __regid__ = 'datacat.clean-resourcefeed-process-logs'
    __select__ = match_user_groups('managers')

    def call(self, eid=None):
        def log_deleted(deleted, resourcefeed_eid):
            self.info('deleted process log files "%s" of ressource feed #%d',
                      ', '.join(map(str, deleted)), resourcefeed_eid)

        if eid is not None:
            deleted = delete_process_logs_for(self._cw, eid)
            log_deleted(deleted, eid)
        else:
            deleted = []
            for eid, in self._cw.find('ResourceFeed'):
                rf_deleted = delete_process_logs_for(self._cw, eid)
                log_deleted(rf_deleted, eid)
                deleted.extend(rf_deleted)
        return deleted



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-datacat geo-spatial entity's classes"""

import json

from rdflib import Literal
from rdflib.namespace import Namespace

from cubicweb.predicates import (
    has_related_entities,
    is_instance,
)
from cubicweb.view import EntityAdapter

from cubes.skos.entities import AbstractRDFAdapter
from cubes.datacat import register_location_rdf_mapping
from cubes.datacat.geoutils import merge_in_geojson_collection
from . import add_literal_to_graph


class GeoJSONLocationAdapter(EntityAdapter):
    """Adapt Location entities to the IGeoJSON interface."""

    __regid__ = 'IGeoJSON'
    __select__ = EntityAdapter.__select__ & is_instance('Location')

    def as_geojson(self):
        """Return a JSON object that is the GeoJSON representation of the location.

        Returned GeoJSON object is a ``Feature`` (or an empty object if the location cannot be
        converted to a valid GeoJSON feature).
        """
        # Compute geometry attribute
        rql = ('Any ST_ASGEOJSON(G) WHERE L geometry G, NOT L geometry NULL, '
               'L eid %(location_eid)s')
        rset = self._cw.execute(rql, {'location_eid': self.entity.eid})
        if not rset:  # A GeoJSON feature must have a geometry attribute
            return {}
        # Compute GeoJSON object
        jsonobj = {
            'type': 'Feature',
            'properties': {
                'name': self.entity.name,
                'html_info': u'<strong>{}</strong>'.format(self.entity.name),
            },
            'geometry': json.loads(rset.rows[0][0]),
        }
        if self.entity.cwuri:  # Optional identifier
            jsonobj['id'] = self.entity.cwuri
        return jsonobj


class GeoJSONDatasetAdapter(EntityAdapter):
    """Adapt Dataset entities to the IGeoJSON interface."""

    __regid__ = 'IGeoJSON'
    __select__ = EntityAdapter.__select__ & has_related_entities('locate_datasets', role='object')

    def as_geojson(self):
        """Return a JSON object that is the GeoJSON representation of the dataset.

        Returned GeoJSON object is a ``FeatureCollection`` (or an empty object if the dataset has
        no related locations).
        """
        # Compute feature list from related locations
        locations = self.entity.related('locate_datasets', role='object', entities=True,
                                        targettypes=('Location',))
        features = (location.cw_adapt_to('IGeoJSON').as_geojson() for location in locations)
        return merge_in_geojson_collection(features)


class LocationRDFAdapter(AbstractRDFAdapter):
    """Adapt Location entities to RDF using DCAT vocabulary."""
    __regid__ = 'RDFPrimary'
    __select__ = AbstractRDFAdapter.__select__ & is_instance('Location')
    register_rdf_mapping = staticmethod(register_location_rdf_mapping)

    def fill(self, graph):
        super(LocationRDFAdapter, self).fill(graph)
        reg = self.registry
        rset = self._cw.execute(
            'Any ST_ASTEXT(G) WHERE X geometry G, NOT X geometry NULL, X eid %(x)s',
            {'x': self.entity.eid},
        )
        if rset:
            wkt = rset[0][0]
            gsp = Namespace('http://www.opengis.net/ont/geosparql#')
            value = Literal(wkt, datatype=gsp.wktLiteral)
            add_literal_to_graph(
                self.entity.cwuri, reg.normalize_uri('locn:geometry'), value,
                graph)



# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-datacat schema"""

from yams.buildobjs import (
    ComputedRelation,
    EntityType,
    Int,
    RelationDefinition,
    RichString,
    String,
    TZDatetime,
)
from yams import constraints
from cubicweb import _
from cubicweb.schema import (
    ERQLExpression,
    RQLConstraint,
    RQLVocabularyConstraint,
    RRQLExpression,
)

from cubes.file.schema import File
from cubes.skos.schema import SKOSSource
from cubes.dataprocessing import schema as dataprocessing


# Until https://www.cubicweb.org/ticket/12437063 gets done.
SKOSSource.__permissions__ = SKOSSource.__permissions__.copy()
SKOSSource.__permissions__['read'] = ('managers', 'users')


def publication_metadata(cls):
    """Class decorator that adds publication metadata to an entity type."""
    cls.add_relation(TZDatetime(description=_('publication date')),
                     name='issued')
    cls.add_relation(TZDatetime(description=_('modification date')),
                     name='modified')
    return cls


def in_scheme_constraint(*scheme_uris):
    qs = ' OR '.join('EXISTS(O in_scheme CS{0:d}, CS{0:d} cwuri "{1}")'.format(
        idx, uri) for idx, uri in enumerate(scheme_uris))
    return RQLVocabularyConstraint(qs)


class scheme_relation(RelationDefinition):
    """Special relation from a concept scheme to a relation type, that may be used to restrict
    possible concept of a particular relation without depending on the scheme's name or other weak
    mecanism.
    """
    __permissions__ = {'read': ('managers', 'users', 'guests'),
                       'add': ('managers',),
                       'delete': ('managers',)}
    subject = 'ConceptScheme'
    object = 'CWRType'
    cardinality = '**'
    constraints = [
        RQLVocabularyConstraint('NOT EXISTS(C theme_taxonomy S)'),
        RQLVocabularyConstraint('O name "license_type"'),
    ]


class Agent(EntityType):
    name = String(required=True, fulltextindexed=True)
    email = String()
    description = _(
        'An agent (eg. person, group, software or physical artifact).')


class publisher_type(RelationDefinition):
    subject = 'Agent'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    description = _("type of publisher (if unspecified agent is not a publisher)")
    constraints = [
        in_scheme_constraint('http://purl.org/adms/publishertype/1.0'),
    ]


@publication_metadata
class DataCatalog(EntityType):
    """A data catalog is a curated collection of metadata about datasets."""
    title = String(required=True, fulltextindexed=True)
    description = RichString(fulltextindexed=True)
    homepage = String(description=_('URL to the catalog home page'))


class language(RelationDefinition):
    subject = 'DataCatalog'
    object = 'Concept'
    description = _('languages used in the catalog')
    constraints = [
        in_scheme_constraint('http://publications.europa.eu/resource/authority/language'),
    ]


class theme_taxonomy(RelationDefinition):
    subject = 'DataCatalog'
    object = 'ConceptScheme'
    description = _(
        'vocabulary or thesaurus containing a list of themes for datasets in this catalog')


class catalog_publisher(RelationDefinition):
    subject = 'DataCatalog'
    object = 'Agent'
    inlined = True
    cardinality = '?*'
    description = _(
        'agent responsible for making the catalog available')


class spatial_coverage(RelationDefinition):
    subject = ('DataCatalog', 'Dataset')
    object = 'Concept'
    cardinality = '**'
    description = _('a geographical area covered by the entity')
    constraints = [
        in_scheme_constraint(
            'http://publications.europa.eu/resource/authority/country',
            'http://publications.europa.eu/resource/authority/place',
            'http://publications.europa.eu/resource/authority/continent',
            'http://sws.geonames.org/',
        ),
    ]


@publication_metadata
class Dataset(EntityType):
    # DCAT attributes.
    identifier = String(fulltextindexed=True)
    title = String(fulltextindexed=True)
    description = RichString(fulltextindexed=True)
    keyword = String(fulltextindexed=True,
                     description=_(
                         'comma-separated list of keywords or tags describing the dataset'))
    landing_page = String(
        description=_('web page providing access to the dataset, its '
                      'distributions and/or additional information'))
    provenance = String(fulltextindexed=True,  # Not in DCAT-AP spec.
                        description=_('source from which this dataset has been produced or '
                                      'extracted'))


class dataset_frequency(RelationDefinition):
    subject = 'Dataset'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    description = _('frequency at which a dataset is updated')
    constraints = [
        in_scheme_constraint(
            'http://publications.europa.eu/resource/authority/frequency'),
    ]


class dataset_type(RelationDefinition):
    subject = 'Dataset'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    description = _('type of dataset')
    constraints = [
        RQLVocabularyConstraint(
            'O in_scheme CS, CS scheme_relation RT, RT name "dataset_type"'),
    ]


class in_catalog(RelationDefinition):
    subject = 'Dataset'
    object = 'DataCatalog'
    cardinality = '1*'
    inlined = True
    description = _('catalog to which the dataset belongs')


class of_dataset(RelationDefinition):
    subject = 'Distribution'
    object = 'Dataset'
    cardinality = '1*'
    composite = 'object'
    inlined = True
    description = _('dataset to which the distribution is related')


class dataset_contact_point(RelationDefinition):
    subject = 'Dataset'
    object = 'Agent'
    cardinality = '?*'
    description = _('contact information that can be used for flagging '
                    'errors in the dataset or sending comments')
    constraints = [RQLVocabularyConstraint('NOT EXISTS(O publisher_type T)')]


class dataset_publisher(RelationDefinition):
    subject = 'Dataset'
    object = 'Agent'
    cardinality = '?*'
    description = _('agent responsible for making the dataset available')
    constraints = [RQLVocabularyConstraint('S in_catalog C, C catalog_publisher O')]


class dataset_contributors(RelationDefinition):
    subject = 'Dataset'
    object = 'Agent'
    cardinality = '**'
    description = _("agent(s) contributing to the dataset's production")


class dcat_theme(RelationDefinition):
    subject = 'Dataset'
    object = 'Concept'
    constraints = [RQLConstraint('O in_scheme CS, S in_catalog C, C theme_taxonomy CS',
                                 msg=_("Theme must belong to a dataset's catalog vocabulary"))]


class license_type(RelationDefinition):
    subject = 'Concept'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    constraints = [
        RQLConstraint('S in_scheme CS, CS scheme_relation RT, RT name "license_type"',
                      msg=_('Only concepts in licenses scheme can have a license type')),
        in_scheme_constraint('http://purl.org/adms/licencetype/1.0'),
    ]


class license(RelationDefinition):
    subject = ('DataCatalog', 'Distribution')
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    composite = 'subject'
    description = _('the licence under which the entity is made available')
    constraints = [RQLVocabularyConstraint('O in_scheme CS, CS scheme_relation RT, '
                                           'RT name "license_type"')]


class catalog_license(ComputedRelation):
    rule = 'S of_dataset D, D in_catalog C, C license O'


class PeriodOfTime(EntityType):
    start = TZDatetime(
        description=_('start of the period'),
    )
    end = TZDatetime(
        description=_('end of the period'),
        constraints=[
            constraints.BoundaryConstraint(
                '>=', constraints.Attribute('start'),
                msg=_('end date must be greater than start date')),
        ],
    )


class temporal_coverage(RelationDefinition):
    subject = 'Dataset'
    object = 'PeriodOfTime'
    cardinality = '*1'
    composite = 'subject'
    description = _('the temporal period that the dataset covers')


@publication_metadata
class Distribution(EntityType):
    title = String(fulltextindexed=True)
    description = RichString(fulltextindexed=True)
    documentation = String(
        description=_('a page or document about this Distribution'))
    access_url = String(
        description=_('A URL that gives access to a distribution of the '
                      'dataset. The resource at the access URL may contain '
                      'information about how to get the dataset.')
    )
    download_url = String(
        description=_('URL to directly download the distribution. '
                      '(Usually unspecified if the distribution is associated '
                      'with a file in the catalog.)')
    )
    byte_size = Int()


class distribution_media_type(RelationDefinition):
    name = 'media_type'
    subject = 'Distribution'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    description = _('The media type of the distribution as defined in the official register of '
                    'media types managed by IANA')


class distribution_format(RelationDefinition):
    subject = 'Distribution'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    description = _('The file format of the distribution')


class file_distribution(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': (RRQLExpression('U has_update_permission O'), ),
        'delete': (RRQLExpression('U has_update_permission O, '
                                  'NOT EXISTS(S produced_by R)'), ),
    }
    subject = 'File'
    object = 'Distribution'
    cardinality = '??'
    inlined = True
    composite = 'object'
    constraints = [
        RQLVocabularyConstraint('NOT EXISTS(SC implemented_by S)'),
    ]


class replaces(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': (),
        'delete': (
            RRQLExpression(
                'U has_update_permission O, U has_update_permission S, '
                'NOT EXISTS(S produced_by P), NOT EXISTS(O produced_by P1)'),
        ),
    }
    subject = 'File'
    object = 'File'
    cardinality = '??'
    inlined = True
    composite = 'subject'
    constraints = [
        RQLConstraint('EXISTS(S file_distribution D)',
                      msg=_('replacement file must be related to a distribution')),
    ]


class ResourceFeed(EntityType):
    """A resource feed is associated to a dataset distribution. It regularly
    fetches files from the specified URL and attached them to this
    distribution.
    """
    __permissions__ = {
        'read': ('managers', ERQLExpression('X owned_by U')),
        'update': (
            'managers',
            ERQLExpression('X owned_by U,'
                           ' NOT EXISTS(X resourcefeed_distribution D, F file_distribution D)')),
        'delete': ('managers', ERQLExpression('X owned_by U')),
        'add': ('managers', 'users')
    }
    title = String(
        description=_('this title will be used to set the title '
                      'of the associated distribution'))
    url = String(required=True)
    processes_to_keep = Int(required=True, default=10,
                            description=_('number of processes data (logs) to keep'))
    versions_to_keep = Int(required=True, default=0,
                           description=_('number of file versions to keep'))
    __unique_toguether__ = [('url', 'resource_feed_of')]


class resourcefeed_media_type(RelationDefinition):
    name = 'media_type'
    subject = 'ResourceFeed'
    object = 'Concept'
    cardinality = '1*'
    inlined = True
    description = _('MIME type of files served by the resource feed.')


class resource_feed_of(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', RRQLExpression('U has_update_permission O')),
        'delete': ('managers', RRQLExpression('U has_update_permission O')),
    }
    subject = 'ResourceFeed'
    object = 'Dataset'
    cardinality = '1*'
    composite = 'object'
    inlined = True


class resourcefeed_distribution(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': (),
        'delete': (),
    }
    subject = 'ResourceFeed'
    object = 'Distribution'
    cardinality = '1*'
    inlined = True


class resource_feed_source(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': (),
        'delete': (),
    }
    subject = 'ResourceFeed'
    object = 'CWSource'
    cardinality = '1?'
    inlined = True
    composite = 'subject'
    # TODO a constraint checking compatibility of data_format between resource
    # feeds sharing the same CWSource.


# Set File permissions:
#  * use `produced_by` relation to prevent modification of generated files
#  * use relative permissions on the Distribution which the file may be a resource
#    of
#  * bind the update permissions on the _Script which uses the File as
#    implementation if any
# XXX Overrides cubicweb-dataprocessing (but we have no obvious way to
# extend ERQLExpression, so...).
_update_file_perms = ('managers', ERQLExpression(
    'NOT EXISTS(X produced_by Y), '
    'NOT EXISTS(X file_distribution D1)'
    ' OR EXISTS(X file_distribution D, U has_update_permission D), '
    'NOT EXISTS(S1 implemented_by X)'
    ' OR EXISTS(S implemented_by X, U has_update_permission S)'
))
File.__permissions__ = File.__permissions__.copy()
File.__permissions__.update({'update': _update_file_perms,
                             'delete': _update_file_perms})


# Customization of dataprocessing for ResourceFeed

dataprocessing.TransformationScript.remove_relation('output_format')

# TransformationSequence is not exposed, restrict its permissions.
dataprocessing.TransformationSequence.__permissions__ = {
    'read': ('managers', 'users'),
    'add': (),
    'update': (),
    'delete': (),
}


class output_media_type(RelationDefinition):
    subject = 'TransformationScript'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    description = _('MIME type of data generated by the script (if unspecified, '
                    'the input file format will be used)')


class script_media_type(RelationDefinition):
    name = 'media_type'
    subject = ('ValidationScript', 'TransformationScript')
    object = 'Concept'
    cardinality = '1*'
    inlined = True
    description = _('MIME type of files handled by the script (use '
                    '"application/octet-stream" for a generic script)')


class _resourcefeed_script(RelationDefinition):
    __abstract__ = True
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', 'users'),
        'delete': (RRQLExpression('U has_update_permission S'), )
    }
    subject = 'ResourceFeed'
    cardinality = '?*'
    inlined = True
    constraints = [
        RQLConstraint(
            'O media_type C, '
            '(L label_of C, L label "application/octet-stream") OR S media_type C',
            msg=_('script does not handle resource feed data format')),
    ]


class validation_script(_resourcefeed_script):
    description = u'a Script used to validated files from this resource'
    object = 'ValidationScript'


class transformation_script(_resourcefeed_script):
    description = u'a Script used to transform files from this resource'
    object = 'TransformationScript'


class process_for_resourcefeed(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', RRQLExpression('U has_update_permission O')),
        'delete': ('managers', RRQLExpression('U has_update_permission O')),
    }
    subject = ('DataTransformationProcess', 'DataValidationProcess')
    object = 'ResourceFeed'
    cardinality = '1*'
    composite = 'object'


dataprocessing.process_depends_on.__permissions__ = {
    'read': ('managers', 'users', 'guests'),
    'add': ('managers', RRQLExpression(
        'NOT EXISTS(S process_for_resourcefeed RF1) '
        'OR (S process_for_resourcefeed RF, U has_update_permission RF)')),
    'delete': ('managers', RRQLExpression(
        'NOT EXISTS(S process_for_resourcefeed RF1) '
        'OR (S process_for_resourcefeed RF, U has_update_permission RF)')),
}


dataprocessing.process_depends_on.constraints = [
    RQLConstraint(
        'NOT EXISTS(S process_for_resourcefeed RF1,'
        '           S process_for_resourcefeed RF2) '
        'OR EXISTS(S process_for_resourcefeed RF,'
        '          O process_for_resourcefeed RF)',
        msg=_('dependency between data process must be within the same '
              'resource feed')),
]


dataprocessing.implemented_by.constraints = [
    RQLVocabularyConstraint(
        'NOT EXISTS(O file_distribution DI, DI of_dataset D)'),
]


dataprocessing.DataTransformationProcess.__permissions__['add'] = ()
dataprocessing.DataTransformationProcess.__permissions__['update'] = ()
dataprocessing.DataValidationProcess.__permissions__['add'] = ()
dataprocessing.DataValidationProcess.__permissions__['update'] = ()



from cubicweb.web.views.urlrewrite import SimpleReqRewriter, rgx


class DistributionReqRewriter(SimpleReqRewriter):
    rules = [
        (rgx('/distribution/(\d+)/raw(/.*)?'),
         dict(rql=r'Any X WHERE X eid \1', vid='download')),
    ]



from logilab.common.shellutils import ProgressBar

from cubes.datacat import cwsource_pull_data, vocabularies

sync_schema_props_perms('resource_feed_source')
sync_schema_props_perms('replaces')

for etype in ('DataCatalog', 'Dataset'):
    for attrtype in ('title', 'description'):
        sync_schema_props_perms((etype, attrtype, 'String'), syncperms=False)
sync_schema_props_perms('catalog_publisher', syncperms=False)

add_attribute('Distribution', 'documentation')
add_attribute('ResourceFeed', 'versions_to_keep')
add_relation_type('dataset_type')

rql('DELETE File F WHERE NOT EXISTS(N replaces F),'
    '                    NOT EXISTS(S implemented_by F),'
    '                    NOT EXISTS(F file_distribution D),'
    '                    NOT EXISTS(P process_stderr F)')
add_entity_type('PeriodOfTime')

add_relation_type('spatial_coverage')

sources = [vocabularies.add_source(cnx, name, url)
           for name, url in vocabularies.DCAT_SCHEMES[-3:]]
commit(ask_confirm=False)
pb = ProgressBar(len(sources),
                 title='synchronizing %d new SKOS sources' % len(sources))
for src in sources:
    cwsource_pull_data(repo, src.through_cw_source[0].eid)
    pb.update()
print('')

add_cubes('leaflet')
add_cubes('postgis')
add_entity_type('Location')



# -*- coding: utf-8 -*-
# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-datacat unit tests for vocabulary processing."""

import json

from rdflib.compare import (
    graph_diff,
    to_isomorphic,
)

from cubicweb.devtools.testlib import BaseTestCase

from cubes.datacat.vocabularies import (
    capitalized_eurovoc_domain,
    eurovoc_domains,
    simplify_eurovoc,
)
from cubes.skos import rdfio


class EurovocTC(BaseTestCase):
    """Test case for Eurovoc manipulation."""

    def test_capitalized_eurovoc_domain(self):
        """Check that Eurovoc domain names case are correctly computed."""
        self.assertEqual(capitalized_eurovoc_domain(u'04 VIE POLITIQUE'), u'04 Vie politique')
        self.assertEqual(capitalized_eurovoc_domain(u'56 AGRICULTURE, SYLVICULTURE ET PÊCHE'),
                         u'56 Agriculture, sylviculture et pêche')
        self.assertEqual(capitalized_eurovoc_domain(u'32 ÉDUCATION ET COMMUNICATION'),
                         u'32 Éducation et communication')

    def test_eurovoc_simplification(self):
        """Check that correct graph is produced from Eurovoc simplification."""
        eurovoc_graph = rdfio.default_graph()
        new_graph = rdfio.default_graph()
        with open(self.datapath('eurovoc_skos.n3')) as f:
            eurovoc_graph.load(f, rdf_format='n3')
        simplify_eurovoc(eurovoc_graph, new_graph)
        expected_graph = rdfio.default_graph()
        with open(self.datapath('eurovoc_skos_simple.n3')) as f:
            expected_graph.load(f, rdf_format='n3')
        iso1 = to_isomorphic(new_graph._graph)
        iso2 = to_isomorphic(expected_graph._graph)
        in_both, in_first, in_second = graph_diff(iso1, iso2)
        self.assertEqual(iso1, iso2,
                         u'RDF graphs are not the same:\n'
                         '* actual:\n{0}\n\n'
                         '* expected:\n{1}'.format(sorted(in_first.serialize(format='n3')),
                                                   sorted(in_second.serialize(format='n3'))))

    def test_eurovoc_domains(self):
        """Check that correct dictionary is produced from Eurovoc domains."""
        simplified_graph = rdfio.default_graph()
        with open(self.datapath('eurovoc_skos_simple.n3')) as f:
            simplified_graph.load(f, rdf_format='n3')
        actual_dict = eurovoc_domains(simplified_graph)
        with open(self.datapath('eurovoc_domains.json')) as f:
            expected_dict = json.load(f)
        self.assertEqual(actual_dict, expected_dict)


if __name__ == '__main__':
    import unittest
    unittest.main()
